#! /usr/env/python
"""
    A python flow accumulation module. It is designed to be general, and to operate across multiple grids and multiple flow direction patterns. However, at the moment, only a steepest descent (single path) routing scheme is implemented.
    
    There remain some outstanding issues with the handling of boundary cells, which this component has inherited from flow_routing_D8.
    
    Created DEJH, 8/2013
"""
try:
    from scipy import weave
except ImportError:
    import warnings
    warnings.warn('unable to import scipy.weave')

import numpy as np
#weave.test()
#import flow_routing_D8

class AccumFlow(object):
    """
    This class allows the routing of flow around a landscape according to a previously calculated flow direction vector. It is not sensitive to grid type. It will eventually be able to work with discharges which are split across more than one node, but at the moment, assumes a single line of descent for a given node.
    """
    def __init__(self, grid, data):
        self.initialize(grid, data)

    def initialize(self, grid, data):
        self.flow_accum_by_area = np.zeros(grid.number_of_nodes+1) #prefilled with zeros, size of WHOLE grid+1, to allow -1 ids
        #Test if this module is to produce the flowacc data. It should be!
        try:
            data.flowacc
        except:
            data.flowacc = self.flow_accum_by_area[:-1]
        else:
            print "Another module has created the flow accumulation grid. Undesirable conflicts may occur!"
            assert(len(data.flowacc) == len(self.flow_accum_by_area[:-1]))

    def calc_flowacc(self, grid, data):
        active_cell_ids = grid.get_active_cell_node_ids()
        #Perform test to see if the flowdir data is a single vector, or multidimensional, here. Several ways possible: 1. Is the vector multidimensional?, e.g., try: data.flowdirs.shape[1] 2. set a flag in flowdir.
        
        try:
            height_order_active_cells = np.argsort(data.elev[active_cell_ids])[::-1] #descending order
        except:
            print 'Cells could not be sorted by elevation. Does the data object contain the elevation vector?'

        try:
            sorted_flowdirs = (data.flowdirs[active_cell_ids])[height_order_active_cells]
        except:
            print 'Flow directions could not be sorted by elevation. Does the data object contain the flow direction vector?'
        #print grid.cell_areas
        self.flow_accum_by_area[active_cell_ids] = grid.cell_areas #This is only the active nodes == cells by definition

        #print len(height_order_active_cells), len(sorted_flowdirs), len(self.flow_accum_by_area)
        #print height_order_active_cells
        #print sorted_flowdirs
        #print data.flowdirs
        #print self.flow_accum_by_area.reshape(5,5)

#---
# Two ways of routing flow are provided. All route flow in descending height order.
#The first, using weave, is not working due to an installation-dependent with the C++ compiler weave uses. However, it will be a massive improvement over other methods
#The second is an inefficient but functional looped method.

        cpp_code_fragment = """
printf ('Test');
"""
        #Shouldn't need to return_val, as we're handling mutable objects not ints

#        flow_accum_by_area = self.flow_accum_by_area
#        a=1.
#        weave.inline(cpp_code_fragment, ['a'], compiler='gcc') #['flow_accum_by_area', 'height_order_active_cells', 'sorted_flowdirs', 'active_cell_ids']) #,verbose=2, compiler='gcc')

#---
        ##inefficient Python code to mimic the above weave:
        for i in xrange(len(sorted_flowdirs)):
            iter_height_order = height_order_active_cells[i]
            iter_sorted_fldirs = sorted_flowdirs[i]
            self.flow_accum_by_area[iter_sorted_fldirs] += (self.flow_accum_by_area[active_cell_ids])[iter_height_order]                     
                                                            
        return self.flow_accum_by_area[:-1]

        #int downhill_node;
        #int active_node;
        #PyObject *active_node_array[(Nactive_cell_ids[0])];
        #for (int i=0; i<(Nactive_cell_ids[0]); i++)
        #    {
        #        active_node_array[i] = &(flow_accum_by_area[(active_cell_ids[i])]);
        #    }
        #
        #for (int i=0; i<(Nsorted_flowdirs[0]); i++)
        #{
        #    downhill_node = sorted_flowdirs[i];
        #    active_node = height_order_active_cells[i];
        #    if downhill_node != active_cell_ids[active_node]
        #    {
        #        flow_accum_by_area[downhill_node] += *(active_node_array[active_node]);
        #    }
        #}



from scipy import weave


def adjust_flow_receivers(fromnode, tonode, elev, link_slope, active_links,
                          receiver, receiver_link, steepest_slope):
    n_nodes = fromnode.size

    code="""
    int f;
    int t;
    for (int i=0; i < n_nodes; i++) {
        f = fromnode[i];
        t = tonode[i];
        if (elev[f] > elev[t] && link_slope[i] > steepest_slope[f]) {
            receiver[f] = t;
            steepest_slope[f] = link_slope[i];
            receiver_link[f] = active_links[i];
        }
        else if (elev[t] > elev[f] && - link_slope[i] > steepest_slope[t]) {
            receiver[t] = f;
            steepest_slope[t] = -link_slope[i];
            receiver_link[t] = active_links[i];
        }
    }
    """

    weave.inline(code, ['n_nodes', 'fromnode', 'tonode', 'elev',
                        'link_slope', 'steepest_slope', 'receiver',
                        'receiver_link', 'active_links'])



#! /usr/bin/env python
"""
Read data from an ESRI ASCII file into a RasterModelGrid
"""

import os
import types
import re

import numpy as np

_VALID_HEADER_KEYS = [
    'ncols', 'nrows', 'xllcorner', 'xllcenter', 'yllcorner',
    'yllcenter', 'cellsize', 'nodata_value',
]
_HEADER_KEY_REGEX_PATTERN = re.compile('\s*(?P<key>[a-zA-z]\w+)')
_HEADER_REGEX_PATTERN = re.compile('\s*(?P<key>[a-zA-Z]\w+)\s+(?P<value>[\w.+-]+)')
_HEADER_VALUE_TESTS = {
    'nrows': (int, lambda x: x > 0),
    'ncols': (int, lambda x: x > 0),
    'cellsize': (float, lambda x: x > 0),
    'xllcorner': (float, lambda x: True),
    'xllcenter': (float, lambda x: True),
    'yllcorner': (float, lambda x: True),
    'yllcenter': (float, lambda x: True),
    'nodata_value': (float, lambda x: True),
}


class Error(Exception):
    """
    Base class for errors in this module.
    """
    pass


class BadHeaderLineError(Error):
    """
    Raise this error if a bad header is line encounter in a ESRI ASCII
    file.
    """
    def __init__(self, line):
        self._line = line

    def __str__(self):
        return self._line


class MissingRequiredKeyError(Error):
    """
    Raise this error when a header is missing a required key.
    """
    def __init__(self, key):
        self._key = key

    def __str__(self):
        return self._key


class KeyTypeError(Error):
    """
    Raise this error when a header's key value is of the wrong type.
    """
    def __init__(self, key, expected_type):
        self._key = key
        self._type = str(expected_type)

    def __str__(self):
        return 'Unable to convert %s to %s' % (self._key, self._type)


class KeyValueError(Error):
    """
    Raise this error when a header's key value has a bad value
    """
    def __init__(self, key, message):
        self._key = key
        self._msg = message

    def __str__(self):
        return '%s: %s' % (self._key, self._msg)


class DataSizeError(Error):
    """
    Raise this error if the size of the data does not match that given in
    the header.
    """
    def __init__(self, size, expected_size):
        self._actual = size
        self._expected = expected_size

    def __str__(self):
        return '%s != %s' % (self._actual, self._expected)


def _parse_header_key_value(line):
    match = _HEADER_KEY_REGEX_PATTERN.match(line)
    if match is None:
        return None
        #raise BadHeaderLineError(line)

    match = _HEADER_REGEX_PATTERN.match(line)
    if match is None:
        raise BadHeaderLineError(line)

    (key, value) = (match.group('key').lower(), match.group('value'))

    if key in _VALID_HEADER_KEYS:
        return (key, value)
    else:
        raise BadHeaderLineError(line)


def _header_lines(asc_file):
    pos = asc_file.tell()
    line = asc_file.readline()
    while len(line) > 0:
        if len(line.strip()) > 0:
            item = _parse_header_key_value(line)
            if item:
                yield item
            else:
                asc_file.seek(pos, 0)
                break
        pos = asc_file.tell()
        line = asc_file.readline()


def _header_is_valid(header):
    header_keys = set(header)
    required_keys = set(['ncols', 'nrows', 'cellsize'])

    try:
        assert(required_keys.issubset(header_keys))
    except AssertionError:
        raise MissingRequiredKeyError(', '.join(required_keys - header_keys))

    for keys in [('xllcenter', 'xllcorner'), ('yllcenter', 'yllcorner')]:
        try:
            assert(len(set(keys) & header_keys) == 1)
        except AssertionError:
            raise MissingRequiredKeyError('|'.join(keys))

    for (key, requires) in _HEADER_VALUE_TESTS.items():
        try:
            header[key] = requires[0](header[key])
            assert(requires[1](header[key]))
        except ValueError:
            raise KeyTypeError(key, float)
        except AssertionError:
            raise KeyValueError(key, 'Bad value')
        except KeyError:
            pass

    return True

def read_asc_header(asc_file):
    """
    Read header information from an ESRI ASCII raster file.

    The header contains the following variables,
        - *ncols*: Number of cell columns
        - *nrows*: Number of cell rows
        - *xllcenter* or *xllcorner*: X (column) coordinate of lower-left
            coordinate of grid (by center or lower-left corner of the cell)
        - *yllcenter*, *yllcorner*: Y (row) coordinate of lower-left
            coordinate of grid (by center or lower-left corner of the cell)
        - *cellsize*: Grid spacing between rows and columns
        - *nodata_value*: No-data value (optional)
    """
    header = dict()
    for (key, value) in _header_lines(asc_file):
        header[key] = value

    try:
        _header_is_valid(header)
    except Error:
        raise

    return header


def _read_asc_data(asc_file):
    """Read gridded data from an ESRI ASCII data file.

    Parameters
    ----------
    asc_file : file-like
        File-like object of the data file pointing to the start of the data.

    .. note::
        First row of the data is at the top of the raster grid, the second
        row is the second from the top, and so on.
    """
    return np.loadtxt(asc_file)


def read_esri_ascii(asc_file, reshape=False, name=None):
    """Read :py:class:`~landlab.RasterModelGrid` from an ESRI ASCII file.

    Read data from *asc_file*, an ESRI_ ASCII file, into a
    :py:class:`~landlab.RasterModelGrid`.  *asc_file* is either the name of
    the data file or is a file-like object.

    The grid and data read from the file are returned as a tuple
    (*grid*, *data*) where *grid* is an instance of
    :py:class:`~landlab.RasterModelGrid` and *data* is a numpy
    array of doubles with that has been reshaped to have the number of rows
    and columns given in the header.

    .. _ESRI: http://resources.esri.com/help/9.3/arcgisengine/java/GP_ToolRef/spatial_analyst_tools/esri_ascii_raster_format.htm

    Parameters
    ----------
    asc_file : str of file-like
        Data file to read.
    reshape : boolean, optional
        Reshape the returned array, otherwise return a flattened array.
    name : str, optional
        Add data to the grid as a named field.

    Returns
    -------
    (grid, data) : tuple
        A newly-created RasterModel grid and the associated node data.
    """
    from ..grid import RasterModelGrid

    if isinstance(asc_file, types.StringTypes):
        file_name = asc_file
        with open(file_name, 'r') as asc_file:
            header = read_asc_header(asc_file)
            data = _read_asc_data(asc_file)
    else:
        header = read_asc_header(asc_file)
        data = _read_asc_data(asc_file)

    shape = (header['nrows'], header['ncols'])
    spacing = (header['cellsize'], header['cellsize'])
    origin = (header['xllcorner'], header['yllcorner'])

    if data.size != shape[0] * shape[1]:
        raise DataSizeError(shape[0] * shape[1], data.size)

    data.shape = shape
    data = np.flipud(data)
    if not reshape:
        data = data.flatten()

    grid = RasterModelGrid(num_rows=shape[0], num_cols=shape[1],
                           dx=spacing[0])
    if name:
        grid.add_field('node', name, data)

    return (grid, data)


def write_esri_ascii(path, fields, names=None, clobber=False):
    """Write landlab fields to ESRI ASCII.

    Write the data and grid information for *fields* to *path* in the ESRI
    ASCII format.

    Parameters
    ----------
    path : str
        Path to output file.
    fields : field-like
        Landlab field object that holds a grid and associated values.
    names : iterable of str, optional
        Names of the fields to include in the output file. If not provided,
        write all fields.
    clobber : boolean
        If *path* exists, clobber the existing file, otherwise raise an
        exception.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.testing.tools import cdtemp
    >>> from landlab import RasterModelGrid
    >>> from landlab.io.esri_ascii import write_esri_ascii

    >>> grid = RasterModelGrid(4, 5, dx=2.)
    >>> _ = grid.add_field('node', 'air__temperature', np.arange(20.))
    >>> with cdtemp() as _:
    ...     files = write_esri_ascii('test.asc', grid)
    >>> files
    ['test.asc']

    >>> _ = grid.add_field('node', 'land_surface__elevation', np.arange(20.))
    >>> with cdtemp() as _:
    ...     files = write_esri_ascii('test.asc', grid)
    >>> files.sort()
    >>> files
    ['test_air__temperature.asc', 'test_land_surface__elevation.asc']
    """
    if os.path.exists(path) and not clobber:
        raise ValueError('file exists')

    if isinstance(names, types.StringTypes):
        names = [names]

    names = names or fields.at_node.keys()

    if len(names) == 1:
        paths = [path]
    elif len(names) > 1:
        (base, ext) = os.path.splitext(path)
        paths = [base + '_' + name + ext for name in names]
    else:
        raise ValueError('no node fields to write')

    bad_names = set(names) - set(fields.at_node.keys())
    if len(bad_names) > 0:
        raise ValueError('unknown field name(s): %s' % ','.join(bad_names))

    header = {
        'ncols': fields.number_of_node_columns,
        'nrows': fields.number_of_node_rows,
        'xllcorner': fields.node_x[0],
        'yllcorner': fields.node_y[0],
        'cellsize': fields.dx,
    }

    for path, name in zip(paths, names):
        with open(path, 'w') as fp:
            for key, val in header.items():
                fp.write('%s %s%s' % (key, str(val), os.linesep))
            data = fields.at_node[name].reshape(header['nrows'],
                                                header['ncols'])
            np.savetxt(fp, np.flipud(data))

    return paths



#! /usr/bin/env python

import numpy

from landlab.grid.voronoi import VoronoiDelaunayGrid


class HexModelGrid(VoronoiDelaunayGrid):
    """A grid of hexagonal cells.

    This inherited class implements a regular 2D grid with hexagonal cells and
    triangular patches. It is a special type of VoronoiDelaunay grid in which
    the initial set of points is arranged in a triangular/hexagonal lattice.
    
    Examples
    --------
    >>> from landlab import HexModelGrid
    >>> hmg = HexModelGrid(3, 2, 1.0)
    >>> hmg.number_of_nodes
    7
    """
    
    def __init__(self, base_num_rows=0, base_num_cols=0, dx=1.0, 
                 orientation='horizontal', shape='hex', reorient_links=False, 
                 **kwds):
        """Create a grid of hexagonal cells.

        Create a regular 2D grid with hexagonal cells and triangular patches.
        It is a special type of VoronoiDelaunay grid in which the initial set
        of points is arranged in a triangular/hexagonal lattice.

        Parameters
        ----------
        base_num_rows : int
            Number of rows of nodes in the left column.
        base_num_cols : int
            Number of nodes on the first row.
        dx : float, optional
            Node spacing.
        orientation : string, optional
            One of the 3 cardinal directions in the grid, either 'horizontal' 
            (default) or 'vertical'

        Returns
        -------
        HexModelGrid
            A newly-created grid.

        Examples
        --------
        Create a hex grid with 2 rows of nodes. The first and third rows will
        have 2 nodes, and the second nodes.

        >>> from landlab import HexModelGrid
        >>> hmg = HexModelGrid(3, 2, 1.0)
        >>> hmg.number_of_nodes
        7
        """
        # Set number of nodes, and initialize if caller has given dimensions
        #self._num_nodes = num_rows * num_cols
        if base_num_rows * base_num_cols > 0:
            self._initialize(base_num_rows, base_num_cols, dx, orientation,
                             shape, reorient_links)
        super(HexModelGrid, self).__init__(**kwds)

    def _initialize(self, base_num_rows, base_num_cols, dx, orientation,
                    shape, reorient_links=False):
        """
        Sets up a hexagonal grid with cell spacing dx and
        (by default) regular boundaries (that is, all perimeter cells are
        boundaries and all interior cells are active).
        
        Parameters
        ----------
        base_num_rows : int
            Number of rows along left side of grid
        base_num_cols : int
            Number of columns along bottom side of grid
        dx : float
            Distance between nodes
        orientation : string
            Either 'horizontal' (default in __init__) or 'vertical'
        shape : string
            Either 'hex' (default in __init__) or 'rect'
        reorient_links : bool
            Whether or not to re-orient all links to point between -45 deg
            and +135 deg clockwise from "north" (i.e., along y axis)

        Returns
        -------
        (none)
        
        Creates/modifies
        ----------------
        Creates and initializes self._num_nodes and self._dx
        
        Notes
        -----
        To be consistent with unstructured grids, the hex grid is
        managed not as a 2D array but rather as a set of arrays that
        describe connectivity information between nodes, links, cells, faces,
        patches, corners, and junctions.
        
        'Horizontal' orientation means that one of the 3 axes of the grid is
        horizontal, whereas the other two are at 30 degree angles to the 
        horizontal, like:
            
            \ /
           -----
            / \
            
        'Vertical' means that one axis is vertical, with the other
        two at 30 degree angles to the vertical, more like:
            
           \   |   /
             \ | /
             / | \
           /   |   \
           
        (of course, these keyboard characters don't represent the angles quite
        right)
        
        Numbers of rows and columns: a hex grid with a rectangular shape will
        have a fixed number of rows and columns, and so for rectangular shaped
        grids we record this information in self._nrows and self._ncols. With
        a hex-shaped grid, either the number of columns (if 'horizontal') or
        the number of rows (if 'vertical') will vary across the grid. Therefore,
        for hex-shaped grids we record only self._nrows for 'horizontal' grids,
        and only self._ncols for 'vertical' grids.
        """
        if self._DEBUG_TRACK_METHODS:
            print 'HexModelGrid._initialize('+str(base_num_rows)+', ' \
                   +str(base_num_cols)+', '+str(dx)+')'
                   
        # Make sure the parameter *orientation* is correct
        assert (orientation[0].lower()=='h' or orientation[0].lower()=='v'), \
               'orientation must be either "horizontal" (default) or "vertical"'
        
        # Make sure the parameter *shape* is correct
        assert (shape[0].lower()=='h' or shape[0].lower()=='r'), \
               'shape must be either "hex" (default) or "rect"'
               
        # Create a set of hexagonally arranged points. These will be our nodes.
        if orientation=='horizontal' and shape=='hex':
            [pts, self._num_nodes] = HexModelGrid.make_hex_points_horizontal_hex(base_num_rows, base_num_cols, dx)
            self.orientation = 'horizontal'
            self._nrows = base_num_rows
        elif orientation=='horizontal' and shape=='rect':
            [pts, self._num_nodes] = HexModelGrid.make_hex_points_horizontal_rect(base_num_rows, base_num_cols, dx)
            self.orientation = 'horizontal'
            self._nrows = base_num_rows
            self._ncols = base_num_cols
        elif orientation=='vertical' and shape=='hex':
            [pts, self._num_nodes] = HexModelGrid.make_hex_points_vertical_hex(base_num_rows, base_num_cols, dx)
            self.orientation = 'vertical'
            self._ncols = base_num_cols
        else:
            [pts, self._num_nodes] = HexModelGrid.make_hex_points_vertical_rect(base_num_rows, base_num_cols, dx)
            self.orientation = 'vertical'
            self._nrows = base_num_rows
            self._ncols = base_num_cols
        
        # Call the VoronoiDelaunayGrid constructor to triangulate/Voronoi
        # the nodes into a grid.
        super(HexModelGrid, self)._initialize(pts[:,0], pts[:,1], reorient_links)
        
        # Remember grid spacing
        self._dx = dx

    def _setup_cell_areas_array(self):
        """
        Creates and returns an array containing the surface areas of the 
        hexagonal (Voronoi) cells.
        
        These cells are perfect hexagons in which the apothem is dx/2. The
        formula for area is:
        
        .. math::
        
            A = 3 dx^2 / 2 \sqrt{3} \approx 0.866 dx^2
        """
        self._cell_areas = 0.8660254*self._dx**2 + numpy.zeros(self.number_of_cells)
        return self._cell_areas
                           
                           
    @staticmethod
    def make_hex_points_horizontal_hex(num_rows, base_num_cols, dxh):
        """
        Creates and returns a set of (x,y) points in a staggered grid in which the 
        points represent the centers of regular hexagonal cells, and the points
        could be connected to form equilateral triangles. The overall shape of the
        lattice is hexagonal, and one of the 3 axes is horizontal.
        
        Inputs: num_rows = number of rows in lattice
                base_num_cols = number of columns in the bottom and top rows
                                (middle rows have more)
                dxh = horizontal and diagonal spacing between points
                
        Return: 2D numpy array containing point (x,y) coordinates, and total number
                of points.
                
        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> [p, npt] = HexModelGrid.make_hex_points_horizontal_hex(3, 2, 1.0)
        >>> npt
        7
        >>> p[1,:]
        array([ 1.,  0.])
        >>> p[:3,0]
        array([ 0. ,  1. , -0.5])
        """

        dxv = dxh * numpy.sqrt(3.) / 2.
        half_dxh = dxh / 2.

        if numpy.mod(num_rows, 2) == 0:  # even number of rows
            npts = num_rows * base_num_cols + (num_rows * num_rows) // 4
            #print 'even # rows, npts=', npts
        else:  # odd number of rows
            npts = num_rows * base_num_cols + ((num_rows - 1) // 2) * ((num_rows - 1) // 2)
            #print 'odd # rows, npts=', npts
        pts = numpy.zeros((npts, 2))
        middle_row = num_rows // 2
        extra_cols = 0
        xshift = 0.
        i = 0
        for r in range(num_rows):
            for c in range(base_num_cols + extra_cols):
                pts[i,0] = c * dxh + xshift
                pts[i,1] = r * dxv
                i += 1
            if r < middle_row:
                extra_cols += 1
            else:
                extra_cols -= 1
            xshift = - half_dxh * extra_cols
        
        return pts, npts


    @staticmethod
    def make_hex_points_horizontal_rect(num_rows, num_cols, dxh):
        """
        Creates and returns a set of (x,y) points in a staggered grid in which the 
        points represent the centers of regular hexagonal cells, and the points
        could be connected to form equilateral triangles. The overall shape of the
        lattice is rectangular, and one of the 3 axes is horizontal.
        
        Inputs: num_rows = number of rows in lattice
                num_cols = number of columns in lattice
                dxh = horizontal and diagonal spacing between points
                
        Return: 2D numpy array containing point (x,y) coordinates, and total number
                of points.
                
        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> [p, npt] = HexModelGrid.make_hex_points_horizontal_rect(3, 3, 1.0)
        >>> npt
        9
        >>> p[1,:]
        array([ 1.,  0.])
        >>> p[:3,0]
        array([ 0.,  1.,  2.])
        """

        dxv = dxh * numpy.sqrt(3.) / 2.
        half_dxh = dxh / 2.

        npts = num_rows * num_cols
        pts = numpy.zeros((npts, 2))
        xshift = 0.
        i = 0
        for r in range(num_rows):
            for c in range(num_cols):
                xshift = half_dxh * (r%2)
                pts[i,0] = c * dxh + xshift
                pts[i,1] = r * dxv
                i += 1
        
        return pts, npts


    @staticmethod
    def make_hex_points_vertical_hex(base_num_rows, num_cols, dxv):
        """
        Creates and returns a set of (x,y) points in a staggered grid in which the 
        points represent the centers of regular hexagonal cells, and the points
        could be connected to form equilateral triangles. The overall shape of the
        lattice is hexagonal.
        
        Inputs: base_num_rows = number of columns in the left and right columns
                                (middle columns have more)
                num_cols = number of columns in lattice
                dxv = vertical and diagonal spacing between points
                
        Return: 2D numpy array containing point (x,y) coordinates, and total number
                of points.
                
        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> [p, npt] = HexModelGrid.make_hex_points_vertical_hex(2, 3, 1.0)
        >>> npt
        7
        >>> p[1,:]
        array([ 0.,  1.])
        >>> p[:3,1]
        array([ 0. ,  1. , -0.5])
        """

        dxh = dxv * numpy.sqrt(3.) / 2.
        half_dxv = dxv / 2.

        if numpy.mod(num_cols, 2) == 0:  # even number of columns
            npts = base_num_rows * num_cols + (num_cols * num_cols) // 4
            #print 'even # cols, npts=', npts
        else:  # odd number of columns
            npts = base_num_rows * num_cols + ((num_cols - 1) // 2) * ((num_cols - 1) // 2)
            #print 'odd # columns, npts=', npts
        pts = numpy.zeros((npts, 2))
        middle_col = num_cols // 2
        extra_rows = 0
        yshift = 0.
        i = 0
        for c in range(num_cols):
            for r in range(base_num_rows + extra_rows):
                pts[i,1] = r * dxv + yshift
                pts[i,0] = c * dxh
                i += 1
            if c < middle_col:
                extra_rows += 1
            else:
                extra_rows -= 1
            yshift = - half_dxv * extra_rows
        
        return pts, npts
        
        
    @staticmethod
    def make_hex_points_vertical_rect(num_rows, num_cols, dxv):
        """
        Creates and returns a set of (x,y) points in a staggered grid in which the 
        points represent the centers of regular hexagonal cells, and the points
        could be connected to form equilateral triangles. The overall shape of the
        lattice is rectangular.
        
        Inputs: num_rows = number of columns in lattice
                num_cols = number of columns in lattice
                dxv = vertical and diagonal spacing between points
                
        Return: 2D numpy array containing point (x,y) coordinates, and total number
                of points.
                
        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> [p, npt] = HexModelGrid.make_hex_points_vertical_rect(3, 3, 1.0)
        >>> npt
        9
        >>> p[1,:]
        array([ 0.,  1.])
        >>> p[:3,1]
        array([ 0.,  1.,  2.])
        """

        dxh = dxv * numpy.sqrt(3.) / 2.
        half_dxv = dxv / 2.

        npts = num_rows * num_cols
        pts = numpy.zeros((npts, 2))
        yshift = 0.
        i = 0
        for c in range(num_cols):
            for r in range(num_rows):
                yshift = half_dxv * (c%2)
                pts[i,1] = r * dxv + yshift
                pts[i,0] = c * dxh
                i += 1
        
        return pts, npts
        
        
    @property
    def number_of_node_columns(self):
        """Number of node columns in a rectangular-shaped and/or
        vertically oriented hex grid.

        Returns the number of columns, including boundaries.
        
        Notes
        -----
        Will generate an error if called with a hex-shaped, horizontally
        aligned grid.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> grid = HexModelGrid(5, 5, shape='rect')
        >>> grid.number_of_node_columns
        5
        """
        return self._ncols


    @property
    def number_of_node_rows(self):
        """Number of node rows in a rectangular-shaped and/or
        horizontally oriented hex grid.

        Returns the number of rows, including boundaries.
        
        Notes
        -----
        Will generate an error if called with a hex-shaped, vertically
        aligned grid.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> grid = HexModelGrid(5, 5, shape='rect')
        >>> grid.number_of_node_rows
        5
        """
        return self._nrows


    def configure_hexplot(self, data, data_label=None, color_map=None):
        """
        Sets up necessary information for making plots of the hexagonal grid
        colored by a given data element.
        
        Parameters
        ----------
        data : str OR node array (1d numpy array with number_of_nodes entries)
            Data field to be colored
        data_label : str, optional
            Label for colorbar
        color_map : matplotlib colormap object, None
            Color map to apply (defaults to "jet")
        
        Returns
        -------
        (none)
        
        Notes
        -----
        Creates and stores a PatchCollection representing the hexagons. Also 
        stores a handle to the current plotting axis. Both of these are then
        used by hexplot().
        """        
        from numpy import array, sqrt, zeros
        import matplotlib
        from matplotlib.patches import Polygon
        from matplotlib.collections import PatchCollection
        #import matplotlib.pyplot as plt
        
        # color
        if color_map is None:
            color_map = matplotlib.cm.jet

        # geometry
        apothem = self._dx/2.0
        radius = 2.0*apothem / sqrt(3.0)  # distance from node to each hexagon cell vertex
        
        # offsets from node x,y position
        offsets = zeros((6,2))
        poly_verts = zeros((6,2))
        
        # Figure out whether the orientation is horizontal or vertical
        if self.node_y[0]==self.node_y[1]:   # horizontal
            offsets[:,0] = array([0., apothem, apothem, 0., -apothem, -apothem])
            offsets[:,1] = array([radius, radius/2.0, -radius/2.0, -radius, -radius/2.0, radius/2.0])
        else:   # vertical
            offsets[:,0] = array([radius/2.0, radius, radius/2.0, -radius/2.0, -radius, -radius/2.0])
            offsets[:,1] = array([apothem, 0., -apothem, -apothem, 0., apothem])
        
        patches = []
        for i in range(self.number_of_nodes):
            poly_verts[:,0] = self.node_x[i]+offsets[:,0]
            poly_verts[:,1] = self.node_y[i]+offsets[:,1]
            p = Polygon(poly_verts, True)
            patches.append(p)
        
        self._hexplot_pc = PatchCollection(patches, cmap=color_map, edgecolor='none', linewidth=0.0)
        
        self._hexplot_configured=True


    def hexplot(self, data, data_label=None, color_map=None):
        """
        Creates a plot of the grid and one node-data field, showing hexagonal
        cells colored by values in the field.
        
        Parameters
        ----------
        data : str OR node array (1d numpy array with number_of_nodes entries)
            Data field to be colored
        data_label : str, optional
            Label for colorbar
        
        Returns
        -------
        (none)
        """
        from numpy import array, amin, amax
        import matplotlib.pyplot as plt
        
        try:
            self._hexplot_configured is True
        except:
            self.configure_hexplot(data, data_label, color_map)

        # Handle *data*: if it's a numpy array, then we consider it the 
        # data to be plotted. If it's a string, we consider it the name of the 
        # node-field to plot, and we fetch it.
        if type(data) is str:
            data_label = data
            data = self.at_node[data]
            
        ax = plt.gca()
        self._hexplot_pc.set_array(array(data))
        ax.add_collection(self._hexplot_pc)
        plt.xlim([amin(self.node_x)-self._dx, amax(self.node_x)+self._dx])
        plt.ylim([amin(self.node_y)-self._dx, amax(self.node_y)+self._dx])
        #cb = plt.colorbar(self._hexplot_pc)
        #if data_label is not None:
        #    cb.set_label(data_label)
        
        #plt.show()


def from_dict(param_dict):
    """
    Create a HexModelGrid from the dictionary-like object, *param_dict*.
    Required keys of the dictionary are NUM_ROWS, NUM_COLS. Raises a KeyError
    if either of these are missing.  If GRID_SPACING is given, use it as the
    HexModelGrid *dx* parameter, otherwise default to unit spacing.
    """
    # Read and create a basic HexModelGrid
    try:
        n_rows = int(param_dict['NUM_ROWS'])
        n_cols = int(param_dict['NUM_COLS'])
        dx = float(param_dict.get('GRID_SPACING', 1.))
    except KeyError as e:
        raise
    except ValueError as e:
        raise
    else:
        hg = HexModelGrid(n_rows, n_cols, dx)
        
    return hg



#! /usr/bin/env python
"""
Read data from a NetCDF file into a RasterModelGrid.
"""

try:
    import netCDF4 as nc4
except ImportError:
    import warnings
    warnings.warn('Unable to import netCDF4.', ImportWarning)

from scipy.io import netcdf as nc

import os
import types
import re

import numpy as np

from landlab.io.netcdf.errors import NotRasterGridError
from landlab.io.netcdf._constants import (_AXIS_DIMENSION_NAMES,
                                          _AXIS_COORDINATE_NAMES,
                                          _COORDINATE_NAMES)


def _length_of_axis_dimension(root, axis_name):
    try:
        return len(root.dimensions[axis_name])
    except TypeError:
        return root.dimensions[axis_name]


def _read_netcdf_grid_shape(root):
    shape = []
    for axis_name in _AXIS_DIMENSION_NAMES:
        try:
            shape.append(_length_of_axis_dimension(root, axis_name))
        except KeyError:
            pass
    return shape


def _read_netcdf_coordinate_values(root):
    values = []
    for coordinate_name in _AXIS_COORDINATE_NAMES:
        try:
            values.append(root.variables[coordinate_name][:].copy())
        except KeyError:
            pass
    return values


def _read_netcdf_coordinate_units(root):
    units = []
    for coordinate_name in _AXIS_COORDINATE_NAMES:
        try:
            units.append(root.variables[coordinate_name].units)
        except KeyError:
            pass
    return units


def _read_netcdf_structured_grid(root):
    shape = _read_netcdf_grid_shape(root)
    coordinates = _read_netcdf_coordinate_values(root)
    units = _read_netcdf_coordinate_units(root)

    for coordinate in coordinates:
        coordinate.shape = shape

    return coordinates


def _read_netcdf_structured_data(root):
    fields = dict()
    for (name, var) in root.variables.items():
        if not name in _COORDINATE_NAMES:
            fields[name] = var[:].copy()
            fields[name].shape = (fields[name].size, )
    return fields


def _get_raster_spacing(coords):
    spacing = np.empty(len(coords), dtype=np.float64)

    for (axis, coord) in enumerate(coords):
        coord_spacing = np.diff(coord, axis=axis)
        try:
            assert(np.all(coord_spacing == coord_spacing.flat[0]))
        except AssertionError:
            raise NotRasterGridError()
        spacing[axis] = coord_spacing.flat[0]

    try:
        assert(np.all(spacing == spacing[0]))
    except AssertionError:
        raise NotRasterGridError()
    else:
        return spacing[0]


def read_netcdf(nc_file, reshape=False, just_grid=False):
    """
    Reads the NetCDF file *nc_file*, and writes it to the fields of a new
    RasterModelGrid, which it then returns.
    Check the names of the fields in the returned grid with
    grid.at_nodes.keys().
    """
    from landlab import RasterModelGrid

    try:
        root = nc.netcdf_file(nc_file, 'r', version=2)
    except TypeError:
        root = nc4.Dataset(nc_file, 'r', format='NETCDF4')

    node_coords = _read_netcdf_structured_grid(root)

    assert(len(node_coords) == 2)

    spacing = _get_raster_spacing(node_coords)

    shape = node_coords[0].shape

    grid = RasterModelGrid(num_rows=shape[0], num_cols=shape[1],
                           dx=spacing)

    if not just_grid:
        fields = _read_netcdf_structured_data(root)
        for (name, values) in fields.items():
            grid.add_field('node', name, values)

    root.close()

    return grid



#! /usr/bin/env python
'''
Plotting functions to extract and plot channel long profiles.
Call all three functions in sequence from the main code.

The functions will return the long profile nodes, return distances upstream of
those nodes, and plot the long profiles, respectively. The former two are, by
necessity, ragged lists of arrays - the number of channels, then the nodes in
that channel.

This module selects channels by taking the largest possible drainages crossing
the grid boundaries. You can specify how many different channels it handles
using the number_of_channels parameter in the channel_nodes function (default 
is 1). This may lead to strange outputs if the drainage structure of the output
changes mid-run (e.g., channel piracy). This may be modified in the future.

DEJH, March 2014.
'''

import numpy
try:
    import matplotlib.pyplot as plt
except ImportError:
    import warnings
    warnings.warn('matplotlib not found', ImportWarning)

def channel_nodes(grid, steepest_nodes, drainage_area, upstream_ID_order, flow_receiver, number_of_channels=1, threshold=None):
    if threshold==None:
        threshold = 2.*numpy.amin(grid.cell_areas)
    boundary_nodes = grid.get_boundary_nodes()
    #top_two_pc = len(boundary_nodes)//50
    #starting_nodes = boundary_nodes[numpy.argsort(drainage_area[boundary_nodes])[-top_two_pc:]]
    starting_nodes = boundary_nodes[numpy.argsort(drainage_area[boundary_nodes])[-number_of_channels:]]
    
    profile_IDs = []
    for i in starting_nodes:
        j=i
        data_store = []
        while 1:
            data_store.append(j)
            supplying_nodes = numpy.where(flow_receiver == j)[0]
            supplying_nodes = supplying_nodes[numpy.where(supplying_nodes != i)]
            max_drainage = numpy.argmax(drainage_area[supplying_nodes])
            if drainage_area[supplying_nodes[max_drainage]] < threshold:
                break
            else:
                j = supplying_nodes[max_drainage]
        profile_IDs.append(numpy.array(data_store))
    return profile_IDs

def get_distances_upstream(grid, len_node_arrays, profile_IDs, links_to_flow_receiver):
    #defined_flow_receivers = numpy.greater_equal(links_to_flow_receiver,-1)
    #flow_link_lengths = numpy.empty(len_node_arrays, dtype=float)
    #flow_link_lengths[defined_flow_receivers] = grid.link_length[links_to_flow_receiver[defined_flow_receivers]]
    #print numpy.sum(defined_flow_receivers)
    distances_upstream = []
    for i in xrange(len(profile_IDs)):
        data_store = []
        total_distance=0.
        data_store.append(total_distance)
        for j in xrange(len(profile_IDs[i])-1):
            total_distance += grid.link_length[links_to_flow_receiver[profile_IDs[i][j+1]]]
            data_store.append(total_distance)
        distances_upstream.append(numpy.array(data_store))
    return distances_upstream

def plot_profiles(distances_upstream, profile_IDs, elevations):
    for i in xrange(len(profile_IDs)):
        the_nodes = profile_IDs[i]
        plt.plot(distances_upstream[i], elevations[the_nodes])
        
def analyze_channel_network_and_plot(grid, elevations='topographic_elevation', 
                                     drainage_area='drainage_area', 
                                     upstream_ID_order='upstream_ID_order', 
                                     flow_receiver='flow_receiver', 
                                     links_to_flow_receiver='links_to_flow_receiver',
                                     number_of_channels=1, 
                                     starting_nodes=None,
                                     threshold=None):
    """
    This function wraps the other three present here, and allows a single-line
    call to plot long profiles.
    As typical elsewhere, the inputs can be field names or arrays.
    
    Note the key new parameter starting_nodes. This (optionally) can be a 
    Python list of node IDs marking the start of each profile. If it is not 
    provided, the profiles with the largest terminal drainage area will be used
    instead.
    
    Returns a tuple, containing:
        - the list of arrays profile_IDs.
        - the list of arrays dists_upstr, the distances from the final, lowest 
            node in the network.
        Both lists are number_of_channels long.
        - 
    """
    internal_list = [0,0,0,0,0] #we're going to put the input arrays in here; must be a better way but this will do
    inputs = (elevations, drainage_area, upstream_ID_order, flow_receiver, links_to_flow_receiver)
    for i in xrange(5):
        j = inputs[i]
        if type(j)==str:
            internal_list[i] = grid.at_node[j]
        else:
            assert j.size == grid.number_of_nodes, "Inputs must be field names or nnode-long numpy arrays!"
            internal_list[i] = j
    
    if starting_nodes==None:
        profile_IDs = channel_nodes(grid, None, internal_list[1], internal_list[2], internal_list[3], number_of_channels, threshold)
    else:
        assert len(starting_nodes)==number_of_channels, "Length of starting_nodes must equal the number_of_channels!"
        if threshold==None:
            threshold = 2.*numpy.amin(grid.cell_areas)
        profile_IDs = []
        for i in starting_nodes:
            j=i
            data_store = []
            while 1:
                data_store.append(j)
                supplying_nodes = numpy.where(flow_receiver == j)[0]
                supplying_nodes = supplying_nodes[numpy.where(supplying_nodes != i)]
                max_drainage = numpy.argmax(internal_list[1][supplying_nodes])
                if internal_list[1][supplying_nodes[max_drainage]] < threshold:
                    break
                else:
                    j = supplying_nodes[max_drainage]
            profile_IDs.append(numpy.array(data_store))
            
    dists_upstr = get_distances_upstream(grid, internal_list[1].size, profile_IDs, internal_list[4])
    plot_profiles(dists_upstr, profile_IDs, internal_list[0])
    
    return (profile_IDs, dists_upstr)



from multiprocessing import Pool

import numpy as np
cimport numpy as np
cimport cython

from libc.math cimport fabs
from libc.stdlib cimport abs


_RHO_MANTLE = 3300.
_GRAVITY = 9.81


DTYPE = np.double
ctypedef np.double_t DTYPE_t


@cython.boundscheck(False)
def subside_parallel_row(np.ndarray[DTYPE_t, ndim=1] w,
                         np.ndarray[DTYPE_t, ndim=1] load,
                         np.ndarray[DTYPE_t, ndim=1] r,
                         alpha):
  cdef int ncols = w.size
  cdef double inv_c = 1. / (2. * np.pi * _RHO_MANTLE * _GRAVITY * alpha ** 2.)
  cdef double c
  cdef int i
  cdef int j

  for i in range(ncols):
    if fabs(load[i]) > 1e-6:
      c = load[i] * inv_c
      for j in range(ncols):
        w[j] += - c * r[abs(j - i)]


def subside_grid(np.ndarray[DTYPE_t, ndim=2] w,
                 np.ndarray[DTYPE_t, ndim=2] load,
                 np.ndarray[DTYPE_t, ndim=2] r,
                 alpha):
  cdef int nrows = w.shape[0]
  cdef i, j

  for i in range(nrows):
    for j in range(nrows):
      subside_parallel_row(w[j], load[i], r[abs(j - i)], alpha)


def subside_grid_strip(np.ndarray[DTYPE_t, ndim=2] load,
                       np.ndarray[DTYPE_t, ndim=2] r,
                       alpha, strip_range):
  (start, stop) = strip_range

  cdef np.ndarray w = np.zeros((stop - start, load.shape[1]), dtype=DTYPE)
  cdef i

  for i in range(load.shape[0]):
    for j in range(start, stop):
      subside_parallel_row(w[j - start], load[i], r[abs(j - i)], alpha)

  return w, strip_range


def tile_grid_into_strips(grid, n_strips):
    rows_per_strip = grid.shape[0] // n_strips

    starts = np.arange(0, grid.shape[0], rows_per_strip)
    stops = starts + rows_per_strip
    stops[-1] = grid.shape[0]

    return zip(starts, stops)


def _subside_grid_strip_helper(args):
  return subside_grid_strip(*args)


def subside_grid_in_parallel(np.ndarray[DTYPE_t, ndim=2] w,
                             np.ndarray[DTYPE_t, ndim=2] load,
                             np.ndarray[DTYPE_t, ndim=2] r,
                             alpha, n_procs):
    if n_procs == 1:
        return subside_grid(w, load, r, alpha)

    strips = tile_grid_into_strips(w, n_procs)

    args = [(load, r, alpha, strip) for strip in strips]

    pool = Pool(processes=n_procs)

    results = pool.map(_subside_grid_strip_helper, args)
    for dz, strip in results:
        start, stop = strip
        w[start:stop] += dz



#################################################################
##
##  Modification of soil_moisture_field.py to accomodate
##  multiple Plant Functional Types (PFTs)
##
##  Sai Nudurupati and Erkan Istanbulluoglu - 31Oct2014
#################################################################

from landlab import Component

import numpy as np

_VALID_METHODS = set(['Grid', 'Multi'])

def assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)

class SoilMoisture( Component ):

    _name = 'Soil Moisture'

    _input_var_names = set([
        'VegetationCover',
        'LiveLeafAreaIndex',
        'PotentialEvapotranspiraton',
    ])

    _output_var_names = set([
        'WaterStress',
        'SaturationFraction',
        'Drainage',
        'Runoff',
        'ActualEvapotranspiration',
    ])

    _var_units = {
        'VegetationCover' : 'None',
        'LiveLeafAreaIndex': 'None',
        'PotentialEvapotranspiraton' : 'mm',
        'WaterStress' : 'Pa',
        'SaturationFraction' : 'None',
        'Drainage' : 'mm',
        'Runoff' : 'mm',
        'ActualEvapotranspiration' : 'mm',
    }


    def __init__( self, grid, data, **kwds ):

        self._method = kwds.pop('method', 'Grid')

        assert_method_is_valid(self._method)

        super(SoilMoisture, self).__init__(grid)

        self.initialize( data, VEGTYPE = grid['cell']['VegetationType'], \
                            **kwds )

        for name in self._input_var_names:
            if not name in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        for name in self._output_var_names:
            if not name in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        self._nodal_values = self.grid['node']

        if not 'InitialSaturationFraction' in self.grid.at_cell:
            self.grid.add_zeros('cell', 'InitialSaturationFraction',
                                    units='None' )

        self._cell_values = self.grid['cell']


    def initialize( self, data, **kwds ):
        # GRASS = 0; SHRUB = 1; TREE = 2; BARE = 3;
        # SHRUBSEEDLING = 4; TREESEEDLING = 5
        self._vegtype = \
          kwds.pop('VEGTYPE', np.zeros(self.grid.number_of_cells,dtype = int))
        self._runon = kwds.pop('RUNON', 0.)
        self._fbare = kwds.pop('F_BARE', 0.7)

        self._interception_cap = \
                np.choose(self._vegtype, kwds.pop('INTERCEPT_CAP',
                [ data['INTERCEPT_CAP_grass'], data['INTERCEPT_CAP_shrub'],
                  data['INTERCEPT_CAP_tree'], data['INTERCEPT_CAP_bare'],
                  data['INTERCEPT_CAP_shrub'], data['INTERCEPT_CAP_tree'] ]))   # Full canopy interception (mm)
        self._zr = np.choose(self._vegtype, kwds.pop('ZR',
                [ data['ZR_grass'], data['ZR_shrub'], data['ZR_tree'],
                  data['ZR_bare'], data['ZR_shrub'], data['ZR_tree'] ]))        # Root depth (m)
        self._soil_Ib = np.choose(self._vegtype, kwds.pop('I_B',
                [ data['I_B_grass'], data['I_B_shrub'], data['I_B_tree'],
                  data['I_B_bare'], data['I_B_shrub'], data['I_B_tree'] ]))     # Infiltration capacity of bare soil (mm/h)
        self._soil_Iv = np.choose(self._vegtype, kwds.pop('I_V',
                [ data['I_V_grass'], data['I_V_shrub'], data['I_V_tree'],
                  data['I_V_bare'], data['I_V_shrub'], data['I_V_tree'] ]))     # Infiltration capacity of vegetated soil (mm/h)
        self._soil_Ew = kwds.pop('EW', [0.1])
        self._soil_pc = np.choose(self._vegtype, kwds.pop('PC',
                [ data['PC_grass'], data['PC_shrub'], data['PC_tree'],
                  data['PC_bare'], data['PC_shrub'], data['PC_tree'] ]))        # Soil porosity
        self._soil_fc = np.choose(self._vegtype, kwds.pop('FC',
                [ data['FC_grass'], data['FC_shrub'], data['FC_tree'],
                  data['FC_bare'], data['FC_shrub'], data['FC_tree'] ]))        # Saturation degree at soil field capacity
        self._soil_sc = np.choose(self._vegtype, kwds.pop('SC',
                [ data['SC_grass'], data['SC_shrub'], data['SC_tree'],
                  data['SC_bare'], data['SC_shrub'], data['SC_tree'] ]))        # Saturation degree at soil stomatal closure
        self._soil_wp = np.choose(self._vegtype, kwds.pop('WP',
                [ data['WP_grass'], data['WP_shrub'], data['WP_tree'],
                  data['WP_bare'], data['WP_shrub'], data['WP_tree'] ]))        # Saturation degree at soil wilting point
        self._soil_hgw = np.choose(self._vegtype, kwds.pop('HGW',
                [ data['HGW_grass'], data['HGW_shrub'], data['HGW_tree'],
                  data['HGW_bare'], data['HGW_shrub'], data['HGW_tree'] ]))     # Saturation degree at soil hygroscopic point
        self._soil_beta = np.choose(self._vegtype, kwds.pop('BETA',
                [ data['BETA_grass'], data['BETA_shrub'], data['BETA_tree'],
                  data['BETA_bare'], data['BETA_shrub'], data['BETA_tree'] ]))  # Deep percolation constant
        self._LAI_max = np.choose( self._vegtype, kwds.pop('LAI_MAX',
                [ data['LAI_MAX_grass'], data['LAI_MAX_shrub'],
                   data['LAI_MAX_tree'], data['LAI_MAX_bare'],
                   data['LAI_MAX_shrub'], data['LAI_MAX_tree'] ]))              # Maximum leaf area index (m2/m2)


    def update( self, current_time, **kwds ):
        #DEBUGG = 0

        P = kwds.pop('P', 5.)
        Tb = kwds.pop('Tb', 24.)
        Tr = kwds.pop('Tr', 0.0)
        self._PET = self._cell_values['PotentialEvapotranspiration']
        self._SO = self._cell_values['InitialSaturationFraction']
        self._vegcover = self._cell_values['VegetationCover']
        self._water_stress = self._cell_values['WaterStress']
        self._S = self._cell_values['SaturationFraction']
        self._D = self._cell_values['Drainage']
        self._ETA = self._cell_values['ActualEvapotranspiration']
        self._fr = self._cell_values['LiveLeafAreaIndex']/self._LAI_max
        self._fr[self._fr > 1.] = 1.
        self._Sini = np.zeros(self._SO.shape)
        self._ETmax = np.zeros(self._SO.shape) # record ETmax - Eq 5 - Zhou et al.


        for cell in range(0,self.grid.number_of_cells):

            #print cell
            s = self._SO[cell]

            fbare = self._fbare
            ZR = self._zr[cell]
            pc = self._soil_pc[cell]
            fc = self._soil_fc[cell]
            sc = self._soil_sc[cell]
            wp = self._soil_wp[cell]
            hgw = self._soil_hgw[cell]
            beta = self._soil_beta[cell]


            Inf_cap = self._soil_Ib[cell]*(1-self._vegcover[cell]) +         \
                                    self._soil_Iv[cell]*self._vegcover[cell]
                                                        # Infiltration capacity
            Int_cap = min(self._vegcover[cell]*self._interception_cap[cell],
                            P*self._vegcover[cell])  # Interception capacity
            Peff = max(P-Int_cap, 0.)         # Effective precipitation depth
            mu = (Inf_cap/1000.0)/(pc*ZR*(np.exp(beta*(1.-fc))-1.))
            Ep = max((self._PET[cell]*self._fr[cell]
                                +fbare*self._PET[cell]*(1.-self._fr[cell]))
                                    - Int_cap, 0.0001)  # mm/d
            self._ETmax[cell] = Ep
            nu = ((Ep/24.)/1000.)/(pc*ZR) # Loss function parameter
            nuw = ((Ep*0.1/24.)/1000.)/(pc*ZR) # Loss function parameter
            sini = self._SO[cell] + ((Peff+self._runon)/(pc*ZR*1000.))

            if sini>1.:
                self._runoff = (sini-1.)*pc*ZR*1000.
                #print 'Runoff =', self._runoff
                sini = 1.

            else:
                self._runoff = 0.


            #self._runon = runoff

            if sini>=fc:
                tfc = (1./(beta*(mu-nu)))*(beta*(fc-sini)+                   \
                        np.log((nu-mu+mu*np.exp(beta*(sini-fc)))/nu))
                tsc = ((fc-sc)/nu)+tfc
                twp = ((sc-wp)/(nu-nuw))*np.log(nu/nuw)+tsc

                if Tb<tfc:
                    s = abs(sini-(1./beta)*np.log(((nu-mu+mu*                 \
                            np.exp(beta*(sini-fc)))*np.exp(beta*(nu-mu)*Tb)   \
                            -mu*np.exp(beta*(sini-fc)))/(nu-mu)))

                    self._D[cell] = ((pc*ZR*1000.)*(sini-s))-(Tb*(Ep/24.))
                    self._ETA[cell] = (Tb*(Ep/24.))

                elif Tb>=tfc and Tb<tsc:
                    s = fc-(nu*(Tb-tfc))
                    self._D[cell] = ((pc*ZR*1000.)*(sini-fc))-((tfc)*(Ep/24.))
                    self._ETA[cell] = (Tb*(Ep/24.))

                elif Tb>=tsc and Tb<twp:
                    s = wp+(sc-wp)*((nu/(nu-nuw))*np.exp((-1)*((nu-nuw)
                                        /(sc-wp))*(Tb-tsc))-(nuw/(nu-nuw)))
                    self._D[cell] = ((pc*ZR*1000.)*(sini-fc))-(tfc*Ep/24.)
                    self._ETA[cell] = (1000.*ZR*pc*(sini-s))-self._D[cell]

                else:
                    s = hgw+(wp-hgw)*np.exp((-1)*(nuw/(wp-hgw))*max(Tb-twp,0.))
                    self._D[cell] = ((pc*ZR*1000.)*(sini-fc))-(tfc*Ep/24.)
                    self._ETA[cell] = (1000.*ZR*pc*(sini-s))-self._D[cell]

            elif sini<fc and sini>=sc:
                tfc = 0.
                tsc = (sini-sc)/nu
                twp = ((sc-wp)/(nu-nuw))*np.log(nu/nuw)+tsc

                if Tb<tsc:
                    s = sini - nu*Tb
                    self._D[cell] = 0.
                    self._ETA[cell] = 1000.*ZR*pc*(sini-s)

                elif Tb>=tsc and Tb<twp:
                    s = wp+(sc-wp)*((nu/(nu-nuw))*np.exp((-1)
                                *((nu-nuw)/(sc-wp))*(Tb-tsc))-(nuw/(nu-nuw)))
                    self._D[cell] = 0
                    self._ETA[cell] = (1000.*ZR*pc*(sini-s))

                else:
                    s = hgw+(wp-hgw)*np.exp((-1)*(nuw/(wp-hgw))*(Tb-twp))
                    self._D[cell] = 0.
                    self._ETA[cell] = (1000.*ZR*pc*(sini-s))

            elif sini<sc and sini>=wp:
                tfc = 0
                tsc = 0
                twp = ((sc-wp)/(nu-nuw))*np.log(1+(nu-nuw)*(sini-wp)
                                        /(nuw*(sc-wp)))

                if Tb<twp:
                    s = wp+((sc-wp)/(nu-nuw))*((np.exp((-1)*((nu-nuw)
                                    /(sc-wp))*Tb))*(nuw+((nu-nuw)
                                        /(sc-wp))*(sini-wp))-nuw)
                    self._D[cell] = 0.
                    self._ETA[cell] = (1000.*ZR*pc*(sini-s))

                else:
                    s = hgw+(wp-hgw)*np.exp((-1)*(nuw/(wp-hgw))*(Tb-twp))
                    self._D[cell] = 0.
                    self._ETA[cell] = (1000.*ZR*pc*(sini-s))

            else:
                tfc = 0.
                tsc = 0.
                twp = 0.

                s = hgw+(sini-hgw)*np.exp((-1)*(nuw/(wp-hgw))*Tb)
                self._D[cell] = 0.
                self._ETA[cell] = (1000.*ZR*pc*(sini-s))

            self._water_stress[cell] = min(((max(((sc - (s+sini)/2.)
                                          /(sc - wp)),0.))**4.),1.0)
            self._S[cell] = s
            self._SO[cell] = s
            self._Sini[cell] = sini

        current_time += (Tb+Tr)/(24.*365.25)
        return( current_time )



#! /usr/env/python
"""
raster_lca.py: simple raster Landlab cellular automaton

This file defines the RasterLCA class, which is a sub-class of 
LandlabCellularAutomaton that implements a simple, non-oriented, raster-grid
CA. Like its parent class, RasterLCA implements a continuous-time, stochastic,
pair-based CA.

Created GT Sep 2014, starting from link_ca.py.
"""

from landlab_ca import LandlabCellularAutomaton, Transition
import landlab


class RasterLCA(LandlabCellularAutomaton):
    
    def __init__(self, model_grid, node_state_dict, transition_list,
                 initial_node_states):
        
        print 'RasterLCA.__init__ here'

        # Make sure caller has sent the right grid type        
        assert (type(model_grid) is landlab.grid.raster.RasterModelGrid), \
               'model_grid must be a Landlab RasterModelGrid'
               
        # Define the number of distinct cell-pair orientations: here just 1,
        # because RasterLCA represents a non-oriented CA model.
        self.number_of_orientations = 1
        
        # Call the LandlabCellularAutomaton.__init__() method to do the rest of
        # the initialization
        super(RasterLCA, self).__init__(model_grid, node_state_dict, 
            transition_list, initial_node_states)
        

if __name__=='__main__':
    print issubclass(RasterLCA, object)
    mg = landlab.RasterModelGrid(3, 4, 1.0)
    nsd = {0 : 'yes', 1 : 'no'}
    xnlist = []
    xnlist.append( Transition( (0,1,0), (1,1,0), 1.0, 'frogging' ) )
    nsg = mg.add_zeros('node', 'node_state_grid')
    rlca = RasterLCA(mg, nsd, xnlist, nsg)
    print rlca.__dict__



#! /usr/bin/env python

import numpy as np
import inspect
from landlab.field.grouped import FieldError
try:
    import matplotlib.pyplot as plt
except ImportError:
    import warnings
    warnings.warn('matplotlib not found', ImportWarning)

from landlab.grid.raster import RasterModelGrid
from landlab.grid.voronoi import VoronoiDelaunayGrid


def assert_array_size_matches(array, size, msg=None):
    if size != array.size:
        if msg is None:
            msg = '%d != %d' % (size, array.size)
        raise ValueError(msg)


def imshow_node_grid(grid, values, **kwds):
    """
    Prepares a map view of data over all nodes in the grid.
    Method can take any of the same **kwds as imshow().
    
    requires:
    grid: the grid
    values: the values on the nodes (length nnodes), or a field name (string)
    from which t draw the data.
    """
    if type(values) == str:
        value_str = values
        values=grid.at_node[values]
    
    assert_array_size_matches(values, grid.number_of_nodes,
            'number of values does not match number of nodes')

    data = values.view()
    
    if RasterModelGrid in inspect.getmro(grid.__class__):
        data.shape = grid.shape

    myimage = _imshow_grid_values(grid, data, **kwds)
    
    try:
        plt.title(value_str)
    except NameError:
        pass
    
    return myimage
    

def imshow_active_node_grid(grid, values, other_node_val='min', **kwds):
    """
    Prepares a map view of data over only the active (i.e., not closed) nodes
    in the grid.
    Method can take any of the same **kwds as imshow().
    
    requires:
    grid: the grid
    values: the values on the open, active nodes OR the values on all nodes in
    the grid. If the latter is provided, this method will only plot the active
    subset.
    
    If *other_node_val* is set, this is the value that will be displayed for
    all nodes that are not active. It defaults to 'min', which is the minimum
    value found on any active node in the grid.
    """
    active_nodes = grid.active_nodes
    try:
        assert_array_size_matches(values, active_nodes.size,
            'number of values does not match number of active nodes')
    except ValueError:
        assert_array_size_matches(values[active_nodes], active_nodes.size,
            'number of values does not match number of active nodes')
        values_to_use = values[active_nodes]
    else:
        values_to_use = values
    
    data = np.zeros(grid.number_of_nodes)
    if other_node_val!='min':
        data.fill(other_node_val)
    else:
        data.fill(np.min(values_to_use))
    data[active_nodes] = values_to_use.flat
    
    if RasterModelGrid in inspect.getmro(grid.__class__):
        data.shape = grid.shape

    myimage = _imshow_grid_values(grid, data, **kwds)
    
    if type(values) == str:
        plt.title(values)
    
    return myimage
    
    
def imshow_core_node_grid(grid, values, other_node_val='min', **kwds):
    """
    Prepares a map view of data over only the core nodes
    in the grid.
    Method can take any of the same **kwds as imshow().
    
    requires:
    grid: the grid
    values: the values on the core nodes OR the values on all nodes in
    the grid. If the latter is provided, this method will only plot the core
    subset. Alternatively, can be a string giving the name of a grid field,
    defined either on core nodes or all nodes.
    
    If *other_node_val* is set, this is the value that will be displayed for
    all nodes that are not core. It defaults to 'min', which is the minimum
    value found on any core node in the grid.
    """
    active_nodes = grid.core_nodes
    
    if type(values) == str:
        value_str = values
        try:
            values=grid.at_core_node[values]
        except FieldError:
            values=grid.at_node[values][active_nodes]
        
    try:
        assert_array_size_matches(values, active_nodes.size,
            'number of values does not match number of active nodes')
    except ValueError:
        assert_array_size_matches(values[active_nodes], active_nodes.size,
            'number of values does not match number of active nodes')
        values_to_use = values[active_nodes]
    else:
        values_to_use = values
    
    data = np.zeros(grid.number_of_nodes)
    if other_node_val!='min':
        data.fill(other_node_val)
    else:
        data.fill(np.min(values_to_use))
    data[active_nodes] = values_to_use.flat
    
    if RasterModelGrid in inspect.getmro(grid.__class__):
        data.shape = grid.shape

    myimage = _imshow_grid_values(grid, data, **kwds)
    
    try:
        plt.title(value_str)
    except NameError:
        pass
    
    return myimage
    

def imshow_cell_grid(grid, values, **kwds):
    """
    Prepares a map view of data over all cells in the grid.
    Method can take any of the same **kwds as imshow().
    
    requires:
    grid: the grid
    values: the values on the cells OR the values on all nodes in the grid, from
    which the cell values will be extracted. Alternatively, can be a field
    name (string) from which to draw the data from the grid.
    """
    cells = grid.node_index_at_cells

    if type(values) == str:
        value_str = values
        try:
            values=grid.at_cell[values]
        except FieldError:
            values=grid.at_node[values][cells]
    
    try:
        assert_array_size_matches(values, cells.size,
            'number of values does not match number of cells')
    except ValueError:
        assert_array_size_matches(values[cells], cells.size,
            'number of values does not match number of cells')
        values_to_use = values[cells]
    else:
        values_to_use = values
        
    data = values_to_use.view()
    if RasterModelGrid in inspect.getmro(grid.__class__):
        data.shape = (grid.shape[0] - 2, grid.shape[1] - 2)

    myimage = _imshow_grid_values(grid, data, **kwds)
    
    try:
        plt.title(value_str)
    except NameError:
        pass
    
    return myimage


def imshow_active_cell_grid(grid, values, other_node_val='min', **kwds):
    """
    Prepares a map view of data over all active (i.e., core and open boundary)
    cells in the grid.
    Method can take any of the same **kwds as imshow().
    
    requires:
    grid: the grid
    values: the values on the active cells OR the values on all nodes in the 
    grid, from which the active cell values will be extracted.
    
    If *other_node_val* is set, this is the value that will be displayed for
    all cells that are not active. It defaults to 'min', which is the minimum
    value found on any active cell in the grid.
    """

    active_cells = grid.node_index_at_active_cells
    
    try:
        assert_array_size_matches(values, active_cells.size,
            'number of values does not match number of active cells')
    except ValueError:
        assert_array_size_matches(values[active_cells], active_cells.size,
            'number of values does not match number of active cells')
        values_to_use = values[active_cells]
    else:
        values_to_use = values

    data = np.zeros(grid.number_of_nodes)
    if other_node_val!='min':
        data.fill(other_node_val)
    else:
        data.fill(np.min(values_to_use))
    data[active_cells] = values_to_use
    data = data[grid.node_index_at_cells]
    if RasterModelGrid in inspect.getmro(grid.__class__):
        data.shape = (grid.shape[0] - 2, grid.shape[1] - 2)

    myimage = _imshow_grid_values(grid, data, **kwds)
    
    return myimage


def _imshow_grid_values(grid, values, var_name=None, var_units=None,
                        grid_units=(None, None), symmetric_cbar=False,
                        cmap='pink', limits=None, allow_colorbar=True,
                        norm=None):
    
    gridtypes = inspect.getmro(grid.__class__)

    if RasterModelGrid in gridtypes:
        if len(values.shape) != 2:
            raise ValueError('dimension of values must be 2 (%s)' % values.shape)
        y = np.arange(values.shape[0] + 1) - grid.dx * .5
        x = np.arange(values.shape[1] + 1) - grid.dx * .5
    
        kwds = dict(cmap=cmap)
        if limits is None:
            if symmetric_cbar:
                (var_min, var_max) = (values.min(), values.max())
                limit = max(abs(var_min), abs(var_max))
                (kwds['vmin'], kwds['vmax']) = (- limit, limit)
        else:
            (kwds['vmin'], kwds['vmax']) = (limits[0], limits[1])
    
    
        myimage = plt.pcolormesh(x, y, values, **kwds)
    
        plt.gca().set_aspect(1.)
        plt.autoscale(tight=True)
        
        if allow_colorbar:
            plt.colorbar(norm=norm)
    
        plt.xlabel('X (%s)' % grid_units[1])
        plt.ylabel('Y (%s)' % grid_units[0])
    
        if var_name is not None:
            plt.title('%s (%s)' % (var_name, var_units))
    
        #plt.show()
        
    elif VoronoiDelaunayGrid in gridtypes:
        # This is still very much ad-hoc, and needs prettifying.
        # We should save the modifications needed to plot color all the way
        # to the diagram edge *into* the grid, for faster plotting.
        # (see http://stackoverflow.com/questions/20515554/colorize-voronoi-diagram)
        # (This technique is not implemented yet)
        from scipy.spatial import voronoi_plot_2d
        import matplotlib.colors as colors
        import matplotlib.cm as cmx
        cm = plt.get_cmap(cmap)
        if limits is None:
            if symmetric_cbar:
                (var_min, var_max) = (values.min(), values.max())
                limit = max(abs(var_min), abs(var_max))
                (vmin, vmax) = (- limit, limit)
            else:
                (vmin, vmax) = (values.min(), values.max())
        else:
            (vmin, vmax) = (limits[0], limits[1])
        cNorm = colors.Normalize(vmin,vmax)
        scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=cm)
        colorVal = scalarMap.to_rgba(values)
        
        myimage = voronoi_plot_2d(grid.vor)
        mycolors = (i for i in colorVal)
        for order in grid.vor.point_region:
            region = grid.vor.regions[order]
            colortouse = next(mycolors)
            if -1 not in region:
                polygon = [grid.vor.vertices[i] for i in region]
                plt.fill(*zip(*polygon),color=colortouse)
                #must be TOTALLY sure the ordering is right

        plt.gca().set_aspect(1.)
        #plt.autoscale(tight=True)
        plt.xlim((np.min(grid.node_x), np.max(grid.node_x)))
        plt.ylim((np.min(grid.node_y), np.max(grid.node_y)))
    
        scalarMap.set_array(values)
        plt.colorbar(scalarMap)
    
        plt.xlabel('X (%s)' % grid_units[1])
        plt.ylabel('Y (%s)' % grid_units[0])
    
        if var_name is not None:
            plt.title('%s (%s)' % (var_name, var_units))
        
    return myimage


def imshow_grid(grid, values, **kwds):
    show = kwds.pop('show', False)
    values_at = kwds.pop('values_at', 'node')

    if values_at == 'node':
        imshow_node_grid(grid, values, **kwds)
    elif values_at == 'cell':
        imshow_cell_grid(grid, values, **kwds)
    elif values_at == 'active_node':
        imshow_active_node_grid(grid, values, **kwds)
    elif values_at == 'active_cell':
        imshow_active_cell_grid(grid, values, **kwds)
    else:
        raise TypeError('value location %s not understood' % values_at)

    if show:
        plt.show()


def imshow_field(field, name, **kwds):
    values_at = kwds['values_at']
    imshow_grid(field, field.field_values(values_at, name), var_name=name,
                var_units=field.field_units(values_at, name), **kwds)

###
# Added by Sai Nudurupati 29Oct2013 
# This function is exactly the same as imshow_grid but this function plots
# arrays spread over cells rather than nodes
##DEJH: Sai, this is duplicating what we already had I think. I deprecated it.
def imshow_active_cells(grid, values, var_name=None, var_units=None,
                grid_units=(None, None), symmetric_cbar=False,
                cmap='pink'):
    """
    .. deprecated:: 0.6
    Use :meth:`imshow_active_cell_grid`, above, instead.
    """
    data = values.view()
    data.shape = (grid.shape[0]-2, grid.shape[1]-2)

    y = np.arange(data.shape[0]) - grid.dx * .5
    x = np.arange(data.shape[1]) - grid.dx * .5

    if symmetric_cbar:
        (var_min, var_max) = (data.min(), data.max())
        limit = max(abs(var_min), abs(var_max))
        limits = (-limit, limit)
    else:
        limits = (None, None)

    plt.pcolormesh(x, y, data, vmin=limits[0], vmax=limits[1], cmap=cmap)

    plt.gca().set_aspect(1.)
    plt.autoscale(tight=True)

    plt.colorbar()

    plt.xlabel('X (%s)' % grid_units[1])
    plt.ylabel('Y (%s)' % grid_units[0])

    if var_name is not None:
        plt.title('%s (%s)' % (var_name, var_units))

    plt.show()

###



#! /usr/bin/env

class Error(Exception):
    """
    Base class for errors in this package.
    """
    pass


class NotRasterGridError(Error):
    """
    Raise this error if the grid defined in the netcdf file is not
    uniform rectilinear with constant spacing in all dimensions.
    """
    pass





#! /usr/bin/env python

import numpy as np


_UNKNOWN_UNITS = '?'


class Error(Exception):
    """Base class for errors in this module."""
    pass


class FieldError(Error, KeyError):
    """Raise this error for a missing field name."""
    def __init__(self, field):
        self._field = field

    def __str__(self):
        return self._field


class ScalarDataFields(dict):
    """Collection of named data fields that are of the same size.

    Holds a collection of data fields that all contain the same number of
    elements and index each of them with a name. This class inherits from
    a standard Python `dict`, which allows access to the fields through
    dict-like syntax.

    Parameters
    ----------
    size : int
        The number of elements in each of the data fields.

    Attributes
    ----------
    units
    size

    See Also
    --------
    landlab.field.ModelDataFields.ones : Hold collections of
        `ScalarDataFields`.
    """
    def __init__(self, size):
        self._size = size

        super(ScalarDataFields, self).__init__()
        self._units = dict()

    @property
    def units(self):
        """Units for values of the field.

        Returns
        -------
        str
            Units of the field.
        """
        return self._units

    @property
    def size(self):
        """Number of elements in the field.

        Returns
        -------
        int
            The number of elements in the field.
        """
        return self._size

    def empty(self, **kwds):
        """Uninitialized array whose size is that of the field.

        Return a new array of the data field size, without initializing
        entries. Keyword arguments are the same as that for the equivalent
        numpy function.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        landlab.field.ScalarDataFields.ones : Equivalent method that
            initializes the data to 1.
        landlab.field.ScalarDataFields.zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> field.empty() # doctest: +SKIP
        array([  2.31584178e+077,  -2.68156175e+154,   9.88131292e-324,
        ... 2.78134232e-309]) # Uninitialized memory

        Note that a new field is *not* added to the collection of fields.

        >>> field.keys()
        []
        """
        return np.empty(self.size, **kwds)

    def ones(self, **kwds):
        """Array, initialized to 1, whose size is that of the field.

        Return a new array of the data field size, filled with ones. Keyword
        arguments are the same as that for the equivalent numpy function.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        landlab.field.ScalarDataFields.empty : Equivalent method that
            does not initialize the new array.
        landlab.field.ScalarDataFields.zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> field.ones()
        array([ 1.,  1.,  1.,  1.])
        >>> field.ones(dtype=int)
        array([1, 1, 1, 1])

        Note that a new field is *not* added to the collection of fields.

        >>> field.keys()
        []
        """
        return np.ones(self.size, **kwds)

    def zeros(self, **kwds):
        """Array, initialized to 0, whose size is that of the field.

        Return a new array of the data field size, filled with zeros. Keyword
        arguments are the same as that for the equivalent numpy function.

        See Also
        --------
        numpy.zeros : See for a description of optional keywords.
        landlab.field.ScalarDataFields.empty : Equivalent method that does not
            initialize the new array.
        landlab.field.scalar_data_fields.ScalarDataFields.ones : Equivalent
            method that initializes the data to 1.

        Examples
        --------
        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> field.zeros()
        array([ 0.,  0.,  0.,  0.])

        Note that a new field is *not* added to the collection of fields.

        >>> field.keys()
        []
        """
        return np.zeros(self.size, **kwds)

    def add_empty(self, name, units=_UNKNOWN_UNITS, **kwds):
        """Create and add an uninitialized array of values to the field.

        Create a new array of the data field size, without initializing
        entries, and add it to the field as *name*. The *units* keyword gives
        the units of the new fields as a string. Remaining keyword arguments
        are the same as that for the equivalent numpy function.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        units : str, optional
            Optionally specify the units of the field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        landlab.field.ScalarDataFields.empty : Equivalent method that
            does not initialize the new array.
        landlab.field.ScalarDataFields.zeros : Equivalent method that
            initializes the data to 0.
        """
        return self.add_field(name, self.empty(**kwds), units=units)

    def add_ones(self, name, units=_UNKNOWN_UNITS, **kwds):
        """Create and add an array of values, initialized to 1, to the field.

        Create a new array of the data field size, filled with ones, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent numpy function.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        units : str, optional
            Optionally specify the units of the field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        andlab.field.ScalarDataFields.add_empty : Equivalent method that
            does not initialize the new array.
        andlab.field.ScalarDataFields.add_zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        Add a new, named field to a collection of fields.

        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> field.add_ones('topographic_elevation')
        array([ 1.,  1.,  1.,  1.])
        >>> field.keys()
        ['topographic_elevation']
        >>> field['topographic_elevation']
        array([ 1.,  1.,  1.,  1.])
        """
        return self.add_field(name, self.ones(**kwds), units=units)

    def add_zeros(self, name, units=_UNKNOWN_UNITS, **kwds):
        """Create and add an array of values, initialized to 0, to the field.

        Create a new array of the data field size, filled with zeros, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent numpy function.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        units : str, optional
            Optionally specify the units of the field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See also
        --------
        numpy.zeros : See for a description of optional keywords.
        landlab.field.ScalarDataFields.add_empty : Equivalent method that
            does not initialize the new array.
        landlab.field.ScalarDataFields.add_ones : Equivalent method that
            initializes the data to 1.
        """
        return self.add_field(name, self.zeros(**kwds), units=units)

    def add_field(self, name, value_array, units=_UNKNOWN_UNITS, copy=False,
                  noclobber=False):
        """Add an array of values to the field.

        Add an array of data values to a collection of fields and associate it
        with the key, *name*. Use the *copy* keyword to, optionally, add a
        copy of the provided array.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        value_array : numpy.array
            Array of values to add to the field.
        units : str, optional
            Optionally specify the units of the field.
        copy : boolean, optional
            If True, add a *copy* of the array to the field. Otherwise save add
            a reference to the array.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        numpy.array
            The data array added to the field. Depending on the *copy*
            keyword, this could be a copy of *value_array* or *value_array*
            itself.

        Raises
        ------
        ValueError :
            If *value_array* has a size different from the field.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.field import ScalarDataFields
        >>> field = ScalarDataFields(4)
        >>> values = np.ones(4, dtype=int)
        >>> field.add_field('topographic_elevation', values)
        array([1, 1, 1, 1])

        A new field is added to the collection of fields. The saved value
        array is the same as the one initially created.

        >>> field['topographic_elevation'] is values
        True

        If you want to save a copy of the array, use the *copy* keyword. In
        addition, adding values to an existing field will remove the reference
        to the previously saved array. The *noclobber* keyword changes this
        behavior to raise an exception in such a case.

        >>> field.add_field('topographic_elevation', values, copy=True)
        array([1, 1, 1, 1])
        >>> field['topographic_elevation'] is values
        False
        >>> field.add_field('topographic_elevation', values, noclobber=True) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        FieldError: topographic_elevation
        """
        if noclobber and name in self:
            raise FieldError(name)

        if copy:
            value_array = value_array.copy()

        self[name] = value_array

        self.set_units(name, units)
        return self[name]

    def set_units(self, name, units):
        """Set the units for a field of values.

        Parameters
        ----------
        name: str
            Name of the field.
        units: str
            Units for the field

        Raises
        ------
        KeyError
            If the named field does not exist.
        """
        self._units[name] = units

    def __setitem__(self, name, value_array):
        if value_array.size != self.size:
            raise ValueError('total size of the new array must be the same as the field')

        if name not in self:
            self.set_units(name, None)

        super(ScalarDataFields, self).__setitem__(name, value_array)

    def __getitem__(self, name):
        try:
            return super(ScalarDataFields, self).__getitem__(name)
        except KeyError:
            raise FieldError(name)



#! /usr/env/python
"""
Link-based cellular automaton modeling tools.

In this updated version of the code, link states can either be encoded by
integer IDs, e.g., 0,1,2..., or tuples actually decribing the state, e.g.,
(x,y,0) to represent a horizontal link between x and y, 
(w,z,1) a vertical link between w and z (w below).
This change remains fully back compatible with GT's original version.

Created GT Oct 2013, modified DEJH Aug 2014 to optionally use identifying pair 
tuples, not just arbitrary IDs.
"""

from heapq import heappush
from heapq import heappop
from landlab import RasterModelGrid
import landlab
import numpy
#from landlab.plot import imshow_grid
import pylab as plt
import time

_NEVER = 1e12

_DEBUG = False

_TEST = False


class Transition():
    """
    Represents a transition from one state ("from_state") to another 
    ("to_state") at a link. The transition probability is represented by a rate 
    parameter "rate", with dimensions of 1/T. The probability distribution of 
    time until the transition event occurs is exponentional with mean 1/rate. 
    The optional name parameter allows the caller to assign a name to any given 
    transition.
    Note that from_state and to_state can now be either integer IDs for the
    standardised ordering of the link states (as before), or tuples explicitly
    describing the node state at each end, and the orientation.
    Orientation is 0: horizontal, L-R; 1: vertical, bottom-top.
    For such a tuple, order is (left/bottom, right/top, orientation).
    """
    def __init__(self, from_state, to_state, rate, name=None):
        
        self.from_state = from_state
        self.to_state = to_state
        self.rate = rate
        self.name = name
        
        
class Event():
    """
    Represents a transition event at a link. The transition occurs at a given
    link and a given time, and it involves a transition into the state xn_to
    (an integer code representing the new link state; "xn" is shorthand for
    "transition").
    
    The class overrides the __lt__ (less than operator) method so that when
    Event() objects are placed in a PriorityQueue, the earliest event is
    given the highest priority (i.e., placed at the top of the queue).
    
    Examples
    --------
    >>> from landlab.components.linkca.link_ca import Event
    >>> e1 = Event( 10.0, 1, 2)
    >>> e2 = Event( 2.0, 3, 1)
    >>> e1 < e2
    False
    >>> e2 < e1
    True
    """
    def __init__(self, time, link, xn_to):

        self.time = time
        self.link = link
        self.xn_to = xn_to
        
    def __lt__(self, other):
        
        return self.time < other.time
            
            
class CAPlotter():
    
    def __init__(self, ca):
        
        self.ca = ca
        
        plt.ion()
        plt.figure(1)
        #nsr=ca.grid.node_vector_to_raster(ca.node_state)
        #plt.imshow(nsr, interpolation='None')
        #plt.draw()
        #plt.pause(0.01)

    def update_plot(self):

        plt.clf()
        nsr = self.ca.grid.node_vector_to_raster(self.ca.node_state)
        plt.imshow(nsr, interpolation='None', origin='lower')
        plt.draw()
        plt.pause(0.01)
        
    def finalize(self):
        
        plt.ioff()
        plt.show()
        
        
class LinkCellularAutomaton():
    """
    The LinkCellularAutomaton implements a link-type (or doublet-type) cellular
    automaton model. A link connects a pair of cells. Each cell has a state
    (represented by an integer code), and each link also has a state that is
    determined by the states of the cell pair.
    """
    def __init__(self, model_grid, node_state_dict, transition_list,
                 initial_node_states, orientation_matters=False):
                 
        # Keep a copy of the model grid
        assert (type(model_grid) is landlab.grid.raster.RasterModelGrid), \
               'model_grid must be a Landlab RasterModelGrid'
        self.grid = model_grid
        self.node_active_links = self.grid.active_node_links()

        self.set_node_state_grid(initial_node_states)
        
        self.current_time = 0.0
        
        #a flag for the creation of link arrays needed to handle grid updates:
        self.translation_index = False
        
        # Figure out how many states there are, and make sure the input data
        # are self consistent.
        #   There are 2 x (N^2) link states, where N is the number of node 
        # states. For example, if there are just two node states, 0 and 1, then 
        # the possible oriented link pairs are listed below:
        #   0-0 0-1 1-0 1-1  0 0 1 1
        #                    0 1 0 1
        self.num_node_states = len(node_state_dict)
        self.num_link_states = 2*self.num_node_states*self.num_node_states  # VARIES WITH LATTICE AND ORIENTATION CHOICE
        
        assert (type(transition_list) is list), 'transition_list must be a list!'
        assert (transition_list), \
               'Transition list must contain at least one transition'
        last_type = None
        for t in transition_list:
            try:
                assert (t.from_state < self.num_link_states), \
                    'Transition from_state out of range'
                assert (t.to_state < self.num_link_states), \
                    'Transition to_state out of range'
                this_type = int
            except: #added to allow from and to states to be tuples, not just ids
                assert type(t.from_state) == tuple, 'Transition from_state out of range'
                assert type(t.to_state) == tuple, 'Transition to_state out of range'
                for i in t.from_state[:-1]:
                    assert (i < self.num_node_states), \
                    'Transition from_state out of range'
                for i in t.to_state[:-1]:
                    assert (i < self.num_node_states), \
                    'Transition to_state out of range'
                assert t.from_state[-1] < 2, \
                    'Encoding for horizontal/vertical in from_state must be 0 or 1.' # VARIES WITH LATTICE AND ORIENTATION CHOICE
                assert t.to_state[-1] < 2, \
                    'Encoding for horizontal/vertical in to_state must be 0 or 1.'
                this_type = tuple
            assert last_type==this_type or last_type==None, \
                'All transition types must be either int IDs, or all tuples.'
            #this test to ensure all entries are either IDs, or tuples, not mixed
            last_type=this_type
        
        # Create priority queue for events and next_update array for links
        self.event_queue = []
        self.next_update = self.grid.create_active_link_array_zeros()
    
        # Assign link types from node types
        self.create_link_state_dict_and_pair_list()

        #DEJH adds: convert transition_list to IDs if necessary
        #This is the new part that allows Transition from_ and to_ types 
        #to be specified either as ints, or as tuples.
        transition_list_as_ID = transition_list[:]
        if type(transition_list[0].from_state) == tuple:
            #(then they all are..., because of the assertions in __init__)
            for i in range(len(transition_list)):
                transition_list_as_ID[i].from_state = self.link_state_dict[transition_list[i].from_state]
                transition_list_as_ID[i].to_state = self.link_state_dict[transition_list[i].to_state]
    
        # Memorize the number of vertical links, so we can figure out
        # node-pair orientations
        # VARIES WITH LATTICE AND ORIENTATION CHOICE
        self.number_of_vertical_links = self.grid.number_of_node_columns * \
                                        (self.grid.number_of_node_rows-1)

        # Using the grid of node states, figure out all the link states
        self.assign_link_states_from_node_types()
    
        # Create transition data for links
        self.setup_transition_data(transition_list_as_ID)

        # Put the various transitions on the event queue
        self.push_transitions_to_event_queue()
        

    def set_node_state_grid(self, node_states):
        """
        Sets the grid of node-state codes to node_states. Also checks
        to make sure node_states is in the proper format, which is to
        say, it's a Numpy array of the same length as the number of nodes in 
        the grid.
        
        Creates: self.node_state (1D Numpy array)
        """
        assert (type(node_states) is numpy.ndarray), \
               'initial_node_states must be a Numpy array'
        assert (len(node_states)==self.grid.number_of_nodes), \
               'length of initial_node_states must equal number of nodes in grid'
        self.grid.at_node['node_state'] = node_states
        self.node_state = node_states
        
                 
    def create_link_state_dict_and_pair_list(self):
        """
        Creates a dictionary that can be used as a lookup table to find out 
        which link state corresponds to a particular pair of node states. The 
        dictionary keys are 3-element tuples, each of which represents the state
        of the FROM node, the TO node, and the orientation of the link. The 
        values are integer codes representing the link state numbers.
        """
        self.link_state_dict = {}
        self.cell_pair = []
        k=0
        for orientation in range(2):   # VARIES WITH LATTICE AND ORIENTATION CHOICE
            for fromstate in range(self.num_node_states):
                for tostate in range(self.num_node_states):
                    self.link_state_dict[(fromstate,tostate,orientation)] = k
                    k+=1
                    self.cell_pair.append((fromstate,tostate,orientation))
    
        if False and _DEBUG:
            print 
            print 'create_link_state_dict_and_pair_list(): dict is:'
            print self.link_state_dict
            print '  and the pair list is:'
            print self.cell_pair


    def active_link_orientation(self, act_link_id):
        """
        Returns 0 if active link *act_link_id* is horizontal (oriented along 
        x-axis), and 1 if it is vertical (oriented along y-axis).
        """
        # VARIES WITH LATTICE AND ORIENTATION CHOICE
        if self.grid.active_links[act_link_id]<self.number_of_vertical_links:
            return 1
        else:
            return 0
    
    
    def assign_link_states_from_node_types(self):
        """
        Assigns a link-state code for each link, and returns a list of these.
        
        Takes lists/arrays of "from" and "to" node IDs for each link, and a 
        dictionary that associates pairs of node states (represented as a 
        3-element tuple, comprising the FROM state, TO state, and orientation) 
        to link states.
        """
        self.link_state = numpy.zeros(self.grid.number_of_active_links,
                                      dtype=int)
    
        for i in range(self.grid.number_of_active_links):
            orientation = self.active_link_orientation(i)  # VARIES WITH LATTICE AND ORIENTATION CHOICE
            node_pair = (self.node_state[self.grid.activelink_fromnode[i]], \
                         self.node_state[self.grid.activelink_tonode[i]], \
                         orientation)
            #print 'node pair:', node_pair, 'dict:', self.link_state_dict[node_pair]
            self.link_state[i] = self.link_state_dict[node_pair]
        
        if False and _DEBUG:
            print 
            print 'assign_link_states_from_node_types(): the link state array is:'
            print self.link_state


    def setup_transition_data(self, xn_list):
        """
        Using the transition list and the number of link states, creates 
        three arrays that collectively contain data on state transitions:
            n_xn: for each link state, contains the number of transitions out of 
                  that state.
            xn_to: 2D array that records, for each link state and each
                   transition, the new state into which the link transitions.
            xn_rate: 2D array that records, for each link state and each
                     transition, the rate (1/time) of the transition.                 
        """
        # First, create an array that stores the number of possible transitions
        # out of each state.
        self.n_xn = numpy.zeros(self.num_link_states, dtype=int)
        for xn in xn_list:
            self.n_xn[xn.from_state] += 1
        
        # Now, create arrays to hold the "to state" and transition rate for each
        # transition. These arrays are dimensioned N x M where N is the number
        # of states, and M is the maximum number of transitions from a single 
        # state (for example if state 3 could transition either to state 1 or 
        # state 4, and the other states only had one or zero possible 
        # transitions, then the maximum would be 2).
        max_transitions = numpy.max(self.n_xn)
        self.xn_to = numpy.zeros((self.num_link_states, max_transitions), dtype=int)
        self.xn_rate = numpy.zeros((self.num_link_states, max_transitions))
    
        #print n_xn, xn_to, xn_rate
    
        # Populate the "to" and "rate" arrays
        self.n_xn[:] = 0  # reset this and then re-do (inefficient but should work)
        for xn in xn_list:
            #print 'from:',xn.from_state,'to:',xn.to_state,'rate:',xn.rate
            from_state = xn.from_state
            self.xn_to[from_state][self.n_xn[from_state]] = xn.to_state
            self.xn_rate[from_state][self.n_xn[from_state]] = xn.rate
            self.n_xn[from_state] += 1
    
        if False and _DEBUG:
            print 
            print 'setup_transition_data():'
            print '  n_xn',self.n_xn
            print '  to:',self.xn_to
            print '  rate:',self.xn_rate
    
    
    def get_next_event(self, link, current_state, current_time):
        """
        Returns the next event for link with ID "link", which is in state
        "current state".
    
        If there is only one potential transition out of the current state, a 
        time for the transition is selected at random from an exponential 
        distribution with rate parameter appropriate for this transition.
    
        If there are more than one potential transitions, a transition time is 
        chosen for each, and the smallest of these applied.
    
        Assumes that there is at least one potential transition from the current
        state.
    
        Inputs: link - ID of the link
                current_state - current state code for the link
                current_time - current time in simulation (i.e., time of event
                                just processed)
            
        Returns: an Event() object containing the time, link ID, and type of the
                next transition event at this link.
        """
        assert (self.n_xn[current_state]>0), \
               'must have at least one potential transition'
    
        #rate = self.xn_rate[current_state][0]
        #rate = xn_rate[current_state][0] * (_SURFACE - ...
    
        # Find next event time for each potential transition
        if self.n_xn[current_state]==1:
            xn = self.xn_to[current_state][0]
            next_time = numpy.random.exponential(1.0/self.xn_rate[current_state][0])
        else:
            next_time = _NEVER
            xn = None
            for i in range(self.n_xn[current_state]):
                this_next = numpy.random.exponential(1.0/self.xn_rate[current_state][i])
                if this_next < next_time:
                    next_time = this_next
                    xn = self.xn_to[current_state][i]
    
        # Create and setup event, and return it
        my_event = Event(next_time+current_time, link, xn)
    
        if _DEBUG:
            print 'get_next_event():'
            print '  next_time:',my_event.time
            print '  link:',my_event.link
            print '  xn_to:',my_event.xn_to
    
        return my_event
    
    
    def push_transitions_to_event_queue(self):
    
        if False and _DEBUG:
            print 'push_transitions_to_event_queue():',self.num_link_states,self.n_xn
        for i in range(self.grid.number_of_active_links):
        
            #print i, self.link_state[i]
            if self.n_xn[self.link_state[i]] > 0:
                #print 'link',i,'has state',self.link_state[i],'and',self.n_xn[self.link_state[i]],'potential transitions'
                event = self.get_next_event(i, self.link_state[i], 0.0)
                heappush(self.event_queue, event)
                self.next_update[i] = event.time
            
            else:
                self.next_update[i] = _NEVER
            
        if True and _DEBUG:
            print '  push_transitions_to_event_queue(): events in queue are now:'
            for e in self.event_queue:
                print '    next_time:',e.time,'link:',e.link,'xn_to:',e.xn_to
            
            
    def update_node_states(self, fromnode, tonode, new_link_state):
        """
        Updates the states of the two nodes in the given link.
        """
    
        # Remember the previous state of each node so we can detect whether the 
        # state has changed
        old_fromnode_state = self.node_state[fromnode]
        old_tonode_state = self.node_state[tonode]
    
        # Change to the new states
        if self.grid.node_status[fromnode]==landlab.grid.base.CORE_NODE:
            self.node_state[fromnode] = self.cell_pair[new_link_state][0]
        if self.grid.node_status[tonode]==landlab.grid.base.CORE_NODE:
            self.node_state[tonode] = self.cell_pair[new_link_state][1]
    
        if _DEBUG:
            print 'update_node_states() for',fromnode,'and',tonode
            print '  fromnode was',old_fromnode_state,'and is now',self.node_state[fromnode]
            print '  tonode was',old_tonode_state,'and is now',self.node_state[tonode]
    
        return self.node_state[fromnode]!=old_fromnode_state, \
               self.node_state[tonode]!=old_tonode_state
           
           
    def update_link_state(self, link, new_link_state, current_time):
        """
        Implements a link transition by updating the current state of the link
        and (if appropriate) choosing the next transition event and pushing it 
        on to the event queue.
    
        Inputs:
            link - ID of the link to update
            new_link_state - code for the new state
            current_time - current time in simulation
        """
        if _DEBUG:
            print
            print 'update_link_state()'
            
        # If the link connects to a boundary, we might have a different state
        # than the one we planned
        fn = self.grid.activelink_fromnode[link]
        tn = self.grid.activelink_tonode[link]
        if _DEBUG:
            print 'fn',fn,'tn',tn,'fnstat',self.grid.node_status[fn],'tnstat',self.grid.node_status[tn]
        if self.grid.node_status[fn]!=landlab.grid.base.CORE_NODE or \
           self.grid.node_status[tn]!=landlab.grid.base.CORE_NODE:
            fns = self.node_state[self.grid.activelink_fromnode[link]]
            tns = self.node_state[self.grid.activelink_tonode[link]]
            orientation = self.active_link_orientation(link)  # VARIES WITH LATTICE AND ORIENTATION CHOICE
            actual_pair = (fns,tns,orientation)
            new_link_state = self.link_state_dict[actual_pair]
            if _DEBUG:
                print '**Boundary: overriding new link state to',new_link_state
            
        self.link_state[link] = new_link_state
        if self.n_xn[new_link_state] > 0:
            event = self.get_next_event(link, new_link_state, current_time)
            heappush(self.event_queue, event)
            self.next_update[link] = event.time
        else:
            self.next_update[link] = _NEVER
            
        if _DEBUG:
            print
            print '  at link',link
            print '  state changed to',self.link_state[link],self.cell_pair[self.link_state[link]]
            print '  update time now',self.next_update[link]
        
            
    def do_transition(self, event, current_time, plot_each_transition=False,
                      plotter=None):
        """
        Implements a state transition. First checks that the transition is still
        valid by comparing the link's next_update time with the corresponding update
        time in the event object.
        
        If the transition is valid, we:
            1) Update the states of the two nodes attached to the link
            2) Update the link's state, choose its next transition, and push it on
            the event queue.
            3) Update the states of the other links attached to the two nodes, 
            choose their next transitions, and push them on the event queue.
            
        Inputs:
            event - Event() object containing the transition data.
            model_grid - ModelGrid() object
            node_state - array of states for each node
            next_update - time of next update for each link
            pair - list of node-state pairs corresponding to each link state
            link_state - array of states for each link
            n_xs - array with number of transitions out of each link state
            eq - event queue
            xn_rate - array with rate of each transition
            node_active_links - list of arrays containing IDs of links connected to
                each node
            ls_dict - dictionary of link-state codes corresponding to each 
                node-state pair
        
        """
    
        if _DEBUG:
            print
            print 'do_transition() for link',event.link
            
        # We'll process the event if its update time matches the one we have 
        # recorded for the link in question. If not, it means that the link has
        # changed state since the event was pushed onto the event queue, and in that
        # case we'll ignore it.
        if event.time == self.next_update[event.link]:
        
            if _DEBUG:
                print '  event time =',event.time
            
            fromnode = self.grid.activelink_fromnode[event.link]
            tonode = self.grid.activelink_tonode[event.link]
            from_changed, to_changed = self.update_node_states(fromnode, tonode, 
                                                          event.xn_to)
            self.update_link_state(event.link, event.xn_to, event.time)

            # Next, when the state of one of the link's nodes changes, we have to
            # update the states of the OTHER links attached to it. This could happen
            # to one or both nodes.
            if from_changed:
                
                if _DEBUG:
                    print '    fromnode has changed state, so updating its links'
            
                for link in self.node_active_links[:,fromnode]:
                    
                    if _DEBUG:
                        print 'f checking link',link
                    if link!=-1 and link!=event.link:
                    
                        this_link_fromnode = self.grid.activelink_fromnode[link]
                        this_link_tonode = self.grid.activelink_tonode[link]
                        orientation = self.active_link_orientation(link)# VARIES WITH LATTICE AND ORIENTATION CHOICE
                        current_pair = (self.node_state[this_link_fromnode], 
                                        self.node_state[this_link_tonode], orientation)
                        new_link_state = self.link_state_dict[current_pair]
                        self.update_link_state(link, new_link_state, event.time)

            if to_changed:
            
                if _DEBUG:
                    print '    tonode has changed state, so updating its links'
            
                for link in self.node_active_links[:,tonode]:
                
                    if _DEBUG:
                        print 't checking link',link
                    if link!=-1 and link!=event.link:
                    
                        this_link_fromnode = self.grid.activelink_fromnode[link]
                        this_link_tonode = self.grid.activelink_tonode[link]
                        orientation = self.active_link_orientation(link)# VARIES WITH LATTICE AND ORIENTATION CHOICE
                        current_pair = (self.node_state[this_link_fromnode], 
                                        self.node_state[this_link_tonode], orientation)
                        new_link_state = self.link_state_dict[current_pair]
                        self.update_link_state(link, new_link_state, event.time)

            if plot_each_transition and (plotter is not None):
                plotter.update_plot()
                
            if _DEBUG:
                n = self.grid.number_of_nodes
                for r in range(self.grid.number_of_node_rows):
                    for c in range(self.grid.number_of_node_columns):
                        n -= 1
                        print '{0:.0f}'.format(self.node_state[n]),
                    print

        elif _DEBUG:
            print '  event time is',event.time,'but update time is', \
                  self.next_update[event.link],'so event will be ignored'
    
    def update_component_data(self, new_node_state_array, nodes_added, translation_vector, changed_uplift=False):
        """
        Call this method to update all data held by the component, if, for
        example, another component or boundary conditions modify the node 
        statuses outside the component between run steps.
        
        This method updates all necessary properties, including both node and
        link states.
        
        *new_node_state_array* is the updated list of node states, which must
        still all be compatible with the state list originally supplied to
        this component.
        *nodes_added* is a list or vector of nodes which have been added to the
        grid to effect uplift or deformation. This might typically be, e.g.,
        bottom_line_nodes[1:-1].
        *translation_vector* is a len 2 tuple describing the node offsets 
        associated with a single increment of deformation. e.g., (0,1) is unit
        uplift. (1,2) could be uplift on a normal fault dipping 60 degrees to
        the left.
        *changed_uplift* is a flag which tells the method to refresh its 
        stored parameters to reflect a step change in the uplift. Set as true
        if you change *nodes_added* or *translation_vector* during a model run.
        """
        self.set_node_state_grid(new_node_state_array)
        self.assign_link_states_from_node_types()
        self.push_transitions_to_event_queue()
        
        #we also need to update next_update
        if changed_uplift:
            self.translation_index = False
        #create the translation old & new lists if necessary (first time this is called)
        if self.translation_index is False:
            nrows = self.grid.number_of_node_rows
            ncols = self.grid.number_of_node_columns
            current_nodes = numpy.array(nodes_added[:])
            next_nodes = numpy.array(nodes_added[:])
            old_nodes = []
            new_nodes = []
            if translation_vector[0]>0:
                hoz_iter = (nrows-1-nodes_added%ncols)//translation_vector[0]
            elif translation_vector[0]<0:
                hoz_iter = (nodes_added%ncols)//-translation_vector[0]
            else:
                hoz_iter = None
            if translation_vector[1]>0:
                vert_iter = (ncols-1-nodes_added//ncols)//translation_vector[1]
            elif translation_vector[1]<0:
                vert_iter = (nodes_added//ncols)//-translation_vector[1]
            else:
                vert_iter = None
            try:
                max_iter = int(max((hoz_iter.max(),vert_iter.max())))
            except AttributeError:
                if vert_iter is not None:
                    max_iter = vert_iter.max()
                    hoz_iter = numpy.ones_like(nodes_added)*numpy.iinfo(int).max
                elif hoz_iter is not None:
                    max_iter = hoz_iter.max()
                    vert_iter = numpy.ones_like(nodes_added)*numpy.iinfo(int).max
                else:
                    raise ValueError('Is uplift vector (0,0)...???')
            #print nodes_added
            #print hoz_iter
            #print vert_iter
            #print max_iter
            for i in xrange(max_iter):
                print "Building the lists... ", i
                on_grid_nodes = numpy.logical_and(
                                            numpy.greater(hoz_iter,i),
                                            numpy.greater(vert_iter,i))
                #print on_grid_nodes
                next_nodes = current_nodes+nrows*translation_vector[1]+translation_vector[0]
                
                if numpy.any(on_grid_nodes):
                    old_nodes.extend(current_nodes[on_grid_nodes])
                    new_nodes.extend(next_nodes[on_grid_nodes])
                    current_nodes[on_grid_nodes] = next_nodes[on_grid_nodes]
                else:
                    break
            #print old_nodes
            old_links = self.grid.node_activelinks(old_nodes)
            new_links = self.grid.node_activelinks(new_nodes)
            print old_links.shape
            print new_links.shape
            self.old_links, unique_index = numpy.unique(old_links, return_index=True)
            self.new_links = new_links.flat[unique_index]
            self.translation_index = True
            self.links_refreshed = numpy.unique(self.grid.node_links(nodes_added))
            self.link_conversion_dict = dict(zip(self.old_links,self.new_links))
        
        self.next_update[self.new_links] = self.next_update[self.old_links]
        self.next_update[self.links_refreshed] = 0.
        for j in self.event_queue:
            try:
                j.link = self.link_conversion_dict[j.link]
            except KeyError:
                pass
        
                  
    def run(self, run_duration, node_state_grid=None, plot_each_transition=False,
            plotter=None):
        
        if node_state_grid is not None:
            self.set_node_state_grid(node_state_grid)
    
        # Continue until we've run out of either time or events
        while self.current_time < run_duration and self.event_queue:
        
            if _DEBUG:
                print 'Current Time = ', self.current_time
        
            # Pick the next transition event from the event queue
            ev = heappop(self.event_queue)
        
            if _DEBUG:
                print 'Event:',ev.time,ev.link,ev.xn_to
        
            self.do_transition(ev, self.current_time, plot_each_transition,
                               plotter)
            
            # Update current time
            self.current_time = ev.time

        
def example_test2():
    
    #from landlab.io.netcdf import write_netcdf
    
    # INITIALIZE

    # User-defined parameters
    nr = 128
    nc = 128
    plot_interval = 0.5
    #next_plot = plot_interval
    run_duration = 50.0
    report_interval = 5.0  # report interval, in real-time seconds
    
    # Initialize real time
    current_real_time = time.time()
    next_report = current_real_time + report_interval

    # Create grid and set up boundaries
    mg = RasterModelGrid(nr, nc, 1.0)
    #mg.set_inactive_boundaries(True, True, True, True)
    
    # Transition data here represent a body of fractured rock, with rock 
    # represented by nodes with state 0, and saprolite (weathered rock)
    # represented by nodes with state 1. Node pairs (links) with 0-1 or 1-0
    # can undergo a transition to 1-1, representing chemical weathering of the
    # rock.
    ns_dict = { 0 : 'air', 1 : 'immobile soil', 2 : 'mobile soil' }
    xn_list = setup_transition_list2()

    # The initial grid represents a domain with half immobile soil, half air
    node_state_grid = mg.add_zeros('node', 'node_state_map', dtype=int)
    if _DEBUG:
        print (numpy.where(mg.node_y<nr/2),)
    (lower_half,) = numpy.where(mg.node_y<nr/2)
    node_state_grid[lower_half] = 1
    
    # Set the left and right boundary conditions
    node_state_grid[mg.left_edge_node_ids()] = 0
    node_state_grid[mg.right_edge_node_ids()] = 0
    
    # Create the CA model
    ca = LinkCellularAutomaton(mg, ns_dict, xn_list, node_state_grid)
    
    print 'INITIALIZING'
    n = ca.grid.number_of_nodes
    if _DEBUG:
        for r in range(ca.grid.number_of_node_rows):
            for c in range(ca.grid.number_of_node_columns):
                n -= 1
                print '{0:.0f}'.format(ca.node_state[n]),
            print
        
    ca_plotter = CAPlotter(ca)
    
    # RUN
    current_time = 0.0
    #time_slice =  0
    #filename = 'soil_ca1-'+str(time_slice).zfill(5)+'.nc'
    #write_netcdf(filename, ca.grid)
    while current_time < run_duration:
        
        # Once in a while, print out simulation and real time to let the user
        # know that the sim is running ok
        current_real_time = time.time()
        if current_real_time >= next_report:
            print 'Current sim time',current_time,'(',100*current_time/run_duration,'%)'
            next_report = current_real_time + report_interval
        
        ca.run(current_time+plot_interval, ca.node_state, 
               plot_each_transition=False) #, plotter=ca_plotter)
        current_time += plot_interval
        if _DEBUG:
            print 'time:',current_time
            print 'ca time:',ca.current_time
        ca_plotter.update_plot()
        #time_slice += 1
        #filename = 'soil_ca1-'+str(time_slice).zfill(5)+'.nc'
        #write_netcdf(filename, ca.grid)
        if _DEBUG:
            n = ca.grid.number_of_nodes
            for r in range(ca.grid.number_of_node_rows):
                for c in range(ca.grid.number_of_node_columns):
                    n -= 1
                    print '{0:.0f}'.format(ca.node_state[n]),
                print
        
        
    # FINALIZE
    
    # Plot
    ca_plotter.finalize()
        
        
def setup_transition_list():
    """
    Creates and returns a list of Transition() objects. This is a "custom"
    function in the sense that any particular application is determined by the
    transition rules that are created here.
    """
    xn_list = []
    
    xn_list.append( Transition(1, 3, 1., 'weathering') ) # rock-sap to sap-sap
    xn_list.append( Transition(2, 3, 1., 'weathering') ) # sap-rock to sap-sap
        
    if False and _DEBUG:
        print
        print 'setup_transition_list(): list has',len(xn_list),'transitions:'
        for t in xn_list:
            print '  From state',t.from_state,'to state',t.to_state,'at rate',t.rate,'called',t.name
        
    return xn_list
    
    
def setup_transition_list2():
    """
    This one is a crude model of "sticky" particles on a gravity-free hill!
    
    There are three cell states: air (A=0), immobile soil (S=1), and mobile
    soil (M=2)
    
    The transition rules are:
        
        1. Mobilization: A-S > A-M (all orientations equal)
        2. Motion: A-M > M-A (")
        3. Sticking: A-M > A-S (")

    Example of link states associated with 3 node states (0, 1, and 2):
    
    Link states with "horizontal" orientation:

    State:    Pair:  
    0            0-0
    1            0-1
    2            0-2
    3            1-0
    4            1-1
    5            1-2
    6            2-0
    7            2-1
    8            2-2

    Link states with "vertical" orientation (the first number in each pair is
    the lower of the two, i.e., the one with the smaller y coordinate):
    
    State:    Pair:  
    9            0^0
    10           0^1
    11           0^2
    12           1^0
    13           1^1
    14           1^2
    15           2^0
    16           2^1
    17           2^2
   
    """
    xn_list = []
    
    xn_list.append( Transition(1, 2, 1., 'mobilization') ) # 
    xn_list.append( Transition(3, 6, 1., 'mobilization') ) # 
    xn_list.append( Transition(10, 11, 1., 'mobilization') ) # 
    xn_list.append( Transition(12, 15, 1., 'mobilization') ) # 
    xn_list.append( Transition(2, 6, 10., 'left motion') )
    xn_list.append( Transition(6, 2, 10., 'right motion') )
    xn_list.append( Transition(11, 15, 1000., 'downward motion') )
    xn_list.append( Transition(15, 11, 1.0, 'upward motion') )    
        
    if False and _DEBUG:
        print
        print 'setup_transition_list2(): list has',len(xn_list),'transitions:'
        for t in xn_list:
            print '  From state',t.from_state,'to state',t.to_state,'at rate',t.rate,'called',t.name
        
    return xn_list
    
    
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    #main()



#! /usr/bin/env python

import numpy

from .voronoi import VoronoiDelaunayGrid


class RadialModelGrid(VoronoiDelaunayGrid):
    """
    This inherited class implements a circular grid in which grid nodes are
    placed at regular radial and semi-regular arc-wise intervals. That is, if
    the radial spacing between "shells" is dr, the nodes are placed around the
    circular shell at regular intervals that get as close as possible to dr.
    The points are then arranged in a Delaunay triangulation with Voronoi
    cells.
    
    Examples
    --------
    >>> from landlab import RadialModelGrid
    >>> omg = RadialModelGrid(num_shells=1, dr=1., origin_x=0., origin_y=0.)
    >>> omg.number_of_nodes
    7
    """
    
    def __init__(self, num_shells=0, dr=1.0, origin_x=0.0, origin_y=0.0, **kwds):
        """Create a circular grid.

        Create a circular grid in which grid nodes are placed at regular
        radial and semi-regular arc-wise intervals. That is, if the radial
        spacing between *shells* is *dr*, the nodes are placed around the
        circular shell at regular intervals that get as close as possible to
        *dr*.  The points are then arranged in a Delaunay triangulation with
        Voronoi cells.
        
        Parameters
        ----------
        num_shells : int
            Number of rings in the grid.
        dr : float, optional
            Radial interval for rings.
        origin_x : float, optional
            x-coordinate of origin node.
        origin_y : float, optional
            y-coordinate of origin node.

        Returns
        -------
        RadialModelGrid
            A newly-created grid.

        Examples
        --------
        A grid with just one ring will have a node at the origin surrounded
        by six other nodes.

        >>> from landlab import RadialModelGrid
        >>> omg = RadialModelGrid(num_shells=1, dr=1., origin_x=0., origin_y=0.)
        >>> omg.number_of_nodes
        7
        >>> omg.number_of_cells
        1

        A second rings will have 13 nodes.

        >>> omg = RadialModelGrid(2)
        >>> omg.number_of_nodes
        20
        """
        # Set number of nodes, and initialize if caller has given dimensions
        #self._num_nodes = num_rows * num_cols
        if num_shells > 0:
            self._initialize(num_shells, dr, origin_x, origin_y)
        super(RadialModelGrid, self).__init__(**kwds)

    def _initialize( self, num_shells, dr, origin_x=0.0, origin_y=0.0):
        if self._DEBUG_TRACK_METHODS:
            print 'RadialModelGrid._initialize('+str(num_shells)+', '+str(dr)+')'
        
        [pts, npts] = self.make_radial_points(num_shells, dr)
        super(RadialModelGrid, self)._initialize(pts[:,0], pts[:,1])
        
        
    def make_radial_points(self, num_shells, dr, origin_x=0.0, origin_y=0.0):
        """
        Creates and returns a set of (x,y) points placed in a series of
        concentric circles around the origin.
        """
        shells = numpy.arange(0, num_shells) + 1
        twopi = 2 * numpy.pi
        n_pts_in_shell = numpy.round(twopi * shells) # number of points in each shell
        dtheta = twopi / n_pts_in_shell
        npts = int(sum(n_pts_in_shell) + 1)
        pts = numpy.zeros((npts, 2))
        r = shells * dr
        startpt = 1
        for i in numpy.arange(0, num_shells):
            theta = (dtheta[i] * numpy.arange(0, n_pts_in_shell[i]) +
                     dtheta[i] / (i + 1))
            pts[startpt:(startpt + int(n_pts_in_shell[i])), 0] = r[i] * numpy.cos(theta)
            pts[startpt:(startpt + int(n_pts_in_shell[i])), 1] = r[i] * numpy.sin(theta)
            startpt += int(n_pts_in_shell[i])
        pts[:,0] += origin_x
        pts[:,1] += origin_y
        
        return pts, npts



#! /usr/bin/env python


import numpy as np


def count_repeated_values(x):
    """
    Count how many times in an array values repeat and where they appear.

    Return a list of length *n* that gives the values and indices of repeated
    values. The first element of the list will be the values and indices of
    all values that appear once or the first time repeated values appear. The
    next element, values that repeat twice or more, and so on. Thus, the
    length of the returned list will be the maximum number that any value is
    repeated in *x*.

    For example, for an array that contains no repeated values, this function
    just returns a copy of *x*, and the indices to each element.
    
    >>> import numpy as np
    >>> from landlab.utils.count_repeats import count_repeated_values
    >>> counts = count_repeated_values(np.array([20, 30, 40]))
    >>> len(counts)
    1
    >>> counts[0]
    (array([20, 30, 40]), array([0, 1, 2]))

    If *x* contains a repeated value, the first element contains all unique
    values along with their indices. For repeated values, return indices to
    their first occurance. The second element contains values and indices to
    values occuring two or more times.

    >>> counts = count_repeated_values(np.array([20, 30, 40, 30]))
    >>> len(counts)
    2
    >>> counts[0]
    (array([20, 30, 40]), array([0, 1, 2]))
    >>> counts[1]
    (array([30]), array([3]))

    """
    counts = []

    (unique_values, unique_inds) = np.unique(x, return_index=True)
    if len(unique_values) > 0:
        x_inds = np.arange(len(x))
        counts.append((unique_values, unique_inds))

        while 1:
            x = np.delete(x, unique_inds)
            x_inds = np.delete(x_inds, unique_inds)
            (unique_values, unique_inds) = np.unique(x, return_index=True)

            if len(unique_values) > 0:
                counts.append((unique_values, x_inds[unique_inds]))
            else:
                break

    return counts



#!/usr/env/python

"""
flow_accum_bw.py:
    
Implementation of Braun & Willett (2012) algorithm for calculating drainage
area and (optionally) water discharge. Assumes each node has only one downstream
receiver. If water discharge is calculated, the result assumes steady flow 
(that is, hydrologic equilibrium).

The main public function is 

    a, q, s = flow_accumulation(r, b)

which takes an array of receiver-node IDs, r (the nodes that "receive" the flow 
from a each node; this array would be returned by the flow_routing component's
calc_flowdirs() method), and an array of baselevel nodes, b. It returns Numpy 
arrays with the drainage area (a) and discharge (q) at each node, along with an 
array (s) that contains the IDs of the nodes in downstream-to-upstream order.

If you simply want the ordered list by itself, use:
    
    s = make_ordered_node_array(r, b)

Created: GT Nov 2013
"""


import numpy
from scipy import weave
from scipy.weave.build_tools import CompileError

class _DrainageStack():
    """
    The _DrainageStack() class implements Braun & Willett's add_to_stack
    function (as a method) and also keeps track of the counter (j) and the
    stack (s). It is used by the make_ordered_node_array() function.
    """
    def __init__(self, delta, D):
        """
        Initializes the index counter j to zero, creates the stack array s,
        and stores references to delta and D.
        """
        self.j = 0
        self.s = numpy.zeros(len(D), dtype=int)
        self.delta = delta
        self.D = D

    def add_to_stack(self, l, use_weave=False):
        """
        Adds node l to the stack and increments the current index (j).
        
        Examples
        --------
        >>> import numpy as np
        >>> from landlab.components.flow_accum.flow_accum_bw import _DrainageStack
        >>> delta = np.array([ 0,  0,  2,  2,  2,  6,  7,  9, 10, 10, 10])
        >>> D = np.array([0, 2, 1, 4, 5, 7, 6, 3, 8, 9])
        >>> ds = _DrainageStack(delta, D)
        >>> ds.add_to_stack(4)
        >>> ds.s
        array([4, 1, 0, 2, 5, 6, 3, 8, 7, 9])
        """
        self.s[self.j] = l
        self.j += 1
        #make some aliases to make the weave faster & better
        delta = self.delta
        D = self.D
        add_it = self.add_to_stack
        delta_l = int(numpy.take(delta,l))
        delta_lplus1 = int(numpy.take(delta,l+1))
        use_weave=False #trip out!
        if use_weave:
            #This weave is NOT YET FUNCTIONAL. The recursive function call is still confusing DEJH. Note the hardcoded trip out above.
            code = """
                int m;
                py::tuple arg(2);
                for (int n = delta_l; n < delta_lplus1; n++) {
                    m = D[n];
                    arg[0] = m;
                    arg[1] = use_weave;
                    if (m != l) {
                        (void) add_it.call(arg);
                    }
                }
            """
            weave.inline(code, ['delta_l', 'delta_lplus1', 'l', 'D', 'add_it', 'use_weave'])
        else:
            for n in xrange(delta_l, delta_lplus1):
                m = self.D[n]
                if m != l:
                    self.add_to_stack(m)
                
                
def _make_number_of_donors_array(r):
    """Number of donors for each node.

    Creates and returns an array containing the number of donors for each node.
    
    Parameters
    ----------
    r : ndarray
        ID of receiver for each node.
    
    Returns
    -------
    ndarray
        Number of donors for each node.
    
    Examples
    --------
    The example below is from Braun and Willett (2012); nd corresponds to their
    d_i in Table 1.
    
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_bw import _make_number_of_donors_array
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8]) - 1
    >>> nd = _make_number_of_donors_array(r)
    >>> nd
    array([0, 2, 0, 0, 4, 1, 2, 1, 0, 0])
    """
    # Vectorized, DEJH, 5/20/14
#    np = len(r)
#    nd = numpy.zeros(np, dtype=int)
#    for i in xrange(np):
#        nd[r[i]] += 1

    nd = numpy.zeros(r.size, dtype=int)
    max_index = numpy.max(r)
    nd[:(max_index + 1)] = numpy.bincount(r)
    return nd
    
    
def _make_delta_array(nd):
    """
    Creates and returns the "delta" array, which is a list containing, for each 
    node, the array index where that node's donor list begins.
    
    Inputs: nd = array containing number of donors for each node
    Returns: delta array
    
    The example below is from Braun and Willett (2012), and represents 
    \delta_i in their Table 1. Here, the numbers are all one less than in their
    table because here we number indices from 0 rather than 1.
    
    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_bw import _make_delta_array
    >>> nd = np.array([0, 2, 0, 0, 4, 1, 2, 1, 0, 0])
    >>> delta = _make_delta_array(nd)
    >>> delta
    array([ 0,  0,  2,  2,  2,  6,  7,  9, 10, 10, 10])
    """
    #np = len(nd)
    #delta = numpy.zeros(np+1, dtype=int)
    #delta[np] = np   # not np+1 as in B&W because here we number from 0
    #for i in xrange(np-1, -1, -1):
    #    delta[i] = delta[i+1] - nd[i]
    #return delta

    #DEJH efficient delooping (only a small gain)
    np = len(nd)
    delta = numpy.zeros(np+1, dtype=int)
    delta.fill(np)
    delta[-2::-1] -= numpy.cumsum(nd[::-1])
    return delta
    
def _make_array_of_donors(r, delta, use_weave=False):
    """
    Creates and returns an array containing the IDs of donors for each node.
    Essentially, the array is a series of lists (not in the Python list object
    sense) of IDs for each node. See Braun & Willett (2012) for details.
    
    The example below is from Braun & Willett (2012), and produces D_i in their
    Table 1 (except that here the ID numbers are one less, because we number
    indices from zero).
    
    Vectorized - inefficiently! - DEJH, 5/20/14
    
    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_bw import _make_array_of_donors
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8])-1
    >>> delta = np.array([ 0,  0,  2,  2,  2,  6,  7,  9, 10, 10, 10])
    >>> D = _make_array_of_donors(r, delta)
    >>> D
    array([0, 2, 1, 4, 5, 7, 6, 3, 8, 9])
    """    
    np = len(r)
    w = numpy.zeros(np, dtype=int)
    D = numpy.zeros(np, dtype=int)
    if use_weave:
        code = """
            int ri;
            for (int i = 0; i < np; i++) {
                ri = r[i];
                D[delta[ri] + w[ri]] = i;
                w[ri] += 1;
            }
        """
        weave.inline(code, ['np', 'r', 'D', 'delta', 'w'])
    else:
        for i in xrange(np):
            ri = r[i]
            D[delta[ri]+w[ri]] = i
            w[ri] += 1
    return D
    
    #DEJH notes that for reasons he's not clear on, this looped version is
    #actually much slower!
    #D = numpy.zeros(np, dtype=int)
    #wri_fin = numpy.bincount(r)
    #wri_fin_nz = wri_fin.nonzero()[0]
    #wri_fin_nz_T = wri_fin_nz.reshape((wri_fin_nz.size,1))
    #logical = numpy.tile(r,(wri_fin_nz.size,1))==wri_fin_nz_T
    #cum_logical = numpy.cumsum(logical, axis=1)
    #wri = numpy.sum(numpy.where(logical, cum_logical-1,0) ,axis=0)
    #D_index = delta[r] + wri
    #D[D_index] = numpy.arange(r.size)
    #return D


def make_ordered_node_array(receiver_nodes, baselevel_nodes, use_weave=False):
    """
    Creates and returns an array of node IDs that is arranged in order from
    downstream to upstream. 
    
    The lack of a leading underscore is meant to signal that this operation
    could be useful outside of this module!
    
    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_bw import make_ordered_node_array
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8])-1
    >>> b = np.array([4])
    >>> s = make_ordered_node_array(r, b)
    >>> s
    array([4, 1, 0, 2, 5, 6, 3, 8, 7, 9])
    """
    nd = _make_number_of_donors_array(receiver_nodes)
    delta = _make_delta_array(nd)
    D = _make_array_of_donors(receiver_nodes, delta)
    dstack = _DrainageStack(delta, D)
    len_bl_nodes = baselevel_nodes.size
    s = numpy.zeros(D.size, dtype=int)
    add_it = dstack.add_to_stack
    for k in baselevel_nodes:
        add_it(k, use_weave)#=False) #don't think this is a bottleneck, so no C++
    return dstack.s
    
    
def find_drainage_area_and_discharge(s, r, node_cell_area=1.0, runoff=1.0,
                                     boundary_nodes=None, use_weave=False):
    """
    Calculates and returns the drainage area and water discharge at each node.
    
    Inputs: s = ordered (downstream to upstream) array of node IDs
            r = array of receiver IDs for each node
            node_cell_area = scalar or numpy array of cell surface areas for
                             each node. If it's an array, must have same length
                             as s (that is, the number of nodes).
            runoff = scalar or numpy array of local runoff rate at each cell
                     (in water depth per time). If it's an array, must have same
                     length as s (that is, the number of nodes).
                     
    Returns: drainage area and discharge as Numpy arrays
    
    Notes:
        - If node_cell_area not given, the output drainage area is equivalent
          to the number of nodes/cells draining through each point, including
          the local node itself.
        - Give node_cell_area as a scalar when using a regular raster grid.
        - If runoff is not given, the discharge returned will be the same as
          drainage area (i.e., drainage area times unit runoff rate).
        - If using an unstructured Landlab grid, make sure that the input
          argument for node_cell_area is the cell area at each NODE rather than
          just at each CELL. This means you need to include entries for the
          perimeter nodes too. They can be zeros.
          
    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_bw import find_drainage_area_and_discharge
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8])-1
    >>> s = np.array([4, 1, 0, 2, 5, 6, 3, 8, 7, 9])
    >>> a, q = find_drainage_area_and_discharge(s, r)
    >>> a
    array([  1.,   3.,   1.,   1.,  10.,   4.,   3.,   2.,   1.,   1.])
    >>> q
    array([  1.,   3.,   1.,   1.,  10.,   4.,   3.,   2.,   1.,   1.])
        
    """
    
    # Number of points
    np = len(s)
    
    # Initialize the drainage_area and discharge arrays. Drainage area starts
    # out as the area of the cell in question, then (unless the cell has no
    # donors) grows from there. Discharge starts out as the cell's local runoff
    # rate times the cell's surface area.
    drainage_area = numpy.zeros(np, dtype=int) + node_cell_area
    discharge = numpy.zeros(np, dtype=int) + node_cell_area*runoff
    
    # Optionally zero out drainage area and discharge at boundary nodes
    if boundary_nodes is not None:
        drainage_area[boundary_nodes] = 0
        discharge[boundary_nodes] = 0
    
    # Iterate backward through the list, which means we work from upstream to
    # downstream.
    num_pts = len(s)
    if use_weave:
        code = """
            int donor;
            int rcvr;
            for (int i = num_pts-1; i > -1; i--) {
                donor = s[i];
                rcvr = r[donor];
                if (donor != rcvr) {
                    drainage_area[rcvr] += drainage_area[donor];
                    discharge[rcvr] += discharge[donor];
                }
            }
        """
        weave.inline(code, ['num_pts', 's', 'r', 'drainage_area', 'discharge'])
    else:
        for i in xrange(np-1, -1, -1):
            donor = s[i]
            recvr = r[donor]
            if donor != recvr:
                drainage_area[recvr] += drainage_area[donor]
                discharge[recvr] += discharge[donor]
    return drainage_area, discharge
    

def flow_accumulation(receiver_nodes, baselevel_nodes, node_cell_area=1.0,
                      runoff_rate=1.0, boundary_nodes=None, use_weave=False):
    """
    Calculates and returns the drainage area and (steady) discharge at each
    node, along with a downstream-to-upstream ordered list (array) of node IDs.
    
    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_bw import flow_accumulation
    >>> r = np.array([2, 5, 2, 7, 5, 5, 6, 5, 7, 8])-1
    >>> b = np.array([4])
    >>> a, q, s = flow_accumulation(r, b)
    >>> a
    array([  1.,   3.,   1.,   1.,  10.,   4.,   3.,   2.,   1.,   1.])
    >>> q
    array([  1.,   3.,   1.,   1.,  10.,   4.,   3.,   2.,   1.,   1.])
    >>> s
    array([4, 1, 0, 2, 5, 6, 3, 8, 7, 9])
    """
    
    s = make_ordered_node_array(receiver_nodes, baselevel_nodes, use_weave)
    #Note that this ordering of s DOES INCLUDE closed nodes. It really shouldn't! 
    #But as we don't have a copy of the grid accessible here, we'll solve this
    #problem as part of route_flow_dn.
    
    a, q = find_drainage_area_and_discharge(s, receiver_nodes, node_cell_area,
                                            runoff_rate, boundary_nodes, 
                                            use_weave)
    
    return a, q, s
    

if __name__ == '__main__':
    import doctest
    doctest.testmod()



import os
import shutil
import tempfile

from nose.tools import assert_true, assert_equal
from distutils.dir_util import mkpath


class cd(object):
    """Context that changes to a new directory.

    Examples
    --------
    >>> import os, tempfile
    >>> from landlab.testing.tools import cd

    Create a temporary directory for testing.

    >>> test_dir = os.path.realpath(tempfile.mkdtemp())

    Withing the context, we're in the new working directory, after exiting
    the context we're back where we started.

    >>> this_dir = os.getcwd()
    >>> with cd(test_dir) as _:
    ...     wdir = os.getcwd()
    >>> test_dir == wdir
    True
    >>> os.getcwd() == this_dir
    True

    If the new working directory does not exists, create it.

    >>> new_dir = os.path.join(test_dir, 'testing.d')
    >>> os.path.exists(new_dir)
    False
    >>> with cd(new_dir) as _:
    ...     wdir = os.getcwd()
    >>> os.path.exists(new_dir)
    True
    >>> wdir == new_dir
    True
    >>> os.getcwd() == this_dir
    True
    """
    def __init__(self, path_to_dir):
        self._dir = path_to_dir

    def __enter__(self):
        self._starting_dir = os.path.abspath(os.getcwd())
        if not os.path.isdir(self._dir):
            mkpath(self._dir)
        os.chdir(self._dir)
        return os.path.abspath(os.getcwd())

    def __exit__(self, ex_type, ex_value, traceback):
        os.chdir(self._starting_dir)


class cdtemp(object):
    """Context that creates and changes to a temporary directory.

    Examples
    --------
    >>> import os
    >>> from landlab.testing.tools import cdtemp

    Change to the newly-created temporary directory after entering the
    context. Upon exiting, remove the temporary directory and return to the
    original working directory.

    >>> this_dir = os.getcwd()
    >>> with cdtemp() as tdir:
    ...     wdir = os.getcwd()
    >>> this_dir == os.getcwd()
    True
    >>> os.path.exists(wdir)
    False
    """
    def __init__(self, **kwds):
        self._kwds = kwds
        self._tmp_dir = None

    def __enter__(self):
        self._starting_dir = os.path.abspath(os.getcwd())
        self._tmp_dir = tempfile.mkdtemp(**self._kwds)
        os.chdir(self._tmp_dir)
        return os.path.abspath(self._tmp_dir)

    def __exit__(self, ex_type, ex_value, traceback):
        os.chdir(self._starting_dir)
        shutil.rmtree(self._tmp_dir)


def assert_is(expr1, expr2, msg=None):
    assert_true(expr1 is expr2, msg=msg)


def assert_is_not(expr1, expr2, msg=None):
    assert_true(expr1 is not expr2, msg=msg)


def assert_is_none(expr, msg=None):
    assert_true(expr is None, msg=msg)


def assert_is_not_none(expr, msg=None):
    assert_true(expr is not None, msg=msg)


def assert_in(first, second, msg=None):
    assert_true(first in second, msg=msg)


def assert_is_not_in(first, second, msg=None):
    assert_true(first not in second, msg=msg)


def assert_is_instance(obj, cls, msg=None):
    assert_true(isinstance(obj, cls), msg=msg)


def assert_not_is_instance(obj, cls, msg=None):
    assert_true(not isinstance(obj, cls), msg=msg)


def assert_list_equal(list1, list2, msg=None):
    assert_true(isinstance(list1, list), msg=msg)
    assert_true(isinstance(list2, list), msg=msg)
    for a, b in zip(list1, list2):
        assert_equal(a, b, msg=msg)


def assert_tuple_equal(tuple1, tuple2, msg=None):
    assert_true(isinstance(tuple1, tuple), msg=msg)
    assert_true(isinstance(tuple1, tuple), msg=msg)
    for a, b in zip(tuple1, tuple2):
        assert_equal(a, b, msg=msg)


def assert_dict_equal(dict1, dict2, msg=None):
    assert_true(isinstance(dict1, dict), msg=msg)
    assert_true(isinstance(dict2, dict), msg=msg)
    assert_equal(dict1, dict2, msg=msg)


def assert_set_equal(set1, set2, msg=None):
    assert_true(isinstance(set1, set), msg=msg)
    assert_true(isinstance(set2, set), msg=msg)
    assert_equal(set1, set2, msg=msg)



#! /usr/bin/env python

import numpy as np

from .structured import StructuredQuadGrid


class RectilinearGrid(StructuredQuadGrid):
    """
    Parameters
    ----------
    coord : tuple
        Coordinates of node rows and node columns.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.structured_quad.rectilinear import RectilinearGrid
    >>> (y, x) = np.arange(4.), np.arange(5.)
    >>> grid = RectilinearGrid((y, x))
    >>> grid.number_of_nodes
    20
    >>> grid.number_of_node_rows
    4
    >>> grid.number_of_node_columns
    5
    >>> grid.corner_nodes
    array([ 0,  4, 15, 19])
    >>> grid.number_of_cells
    6
    >>> grid.node_row_coord
    array([ 0.,  1.,  2.,  3.])
    >>> grid.node_col_coord
    array([ 0.,  1.,  2.,  3.,  4.])
    """
    def __init__(self, coord):
        """
        Parameters
        ----------
        coord : tuple
            Coordinates of node rows and node columns.
        """
        if len(coord) != 2:
            raise ValueError('only 2d grids are supported')

        shape = (
            len(coord[0]),
            len(coord[1]),
        )
        node_coord = np.meshgrid(*coord, indexing='ij')

        super(RectilinearGrid, self).__init__(node_coord, shape, cells=True)

        self._coord = (coord[0], coord[1])

    @property
    def coord(self):
        """Node row and column coordinates.
        """
        return self._coord

    @property
    def node_row_coord(self):
        return self._coord[0]

    @property
    def node_col_coord(self):
        return self._coord[1]


class UniformRectilinearGrid(RectilinearGrid):
    """
    Examples
    --------
    >>> from landlab.grid.structured_quad.rectilinear import UniformRectilinearGrid
    >>> grid = UniformRectilinearGrid((4, 5), spacing=(2, 3), origin=(-1, 1))
    >>> grid.number_of_nodes
    20

    #>>> grid.number_of_core_nodes
    6
    >>> grid.number_of_node_rows
    4
    >>> grid.number_of_node_columns
    5
    >>> grid.corner_nodes
    array([ 0,  4, 15, 19])
    >>> grid.number_of_cells
    6
    >>> grid.node_row_coord
    array([-1.,  1.,  3.,  5.])
    >>> grid.node_col_coord
    array([  1.,   4.,   7.,  10.,  13.])
    """
    def __init__(self, shape, spacing=(1., 1.), origin=(0., 0.)):
        """
        Parameters
        ----------
        shape : tuple
            Shape of the grid in nodes.
        spacing : tuple, optional
            Spacing between rows and columns.
        origin : tuple, optional
            Coordinates of grid origin.
        """
        if len(shape) != 2:
            raise ValueError('only 2d grids are supported')

        coords = (
            np.arange(origin[0], origin[0] + shape[0] * spacing[0], spacing[0],
                      dtype=np.float64),
            np.arange(origin[1], origin[1] + shape[1] * spacing[1], spacing[1],
                      dtype=np.float64),
        )

        super(UniformRectilinearGrid, self).__init__(coords)

        self._spacing = tuple(spacing)

    @property
    def spacing(self):
        """Spacing of rows and columns of grid nodes.
        """
        return self._spacing

    @property
    def dy(self):
        """Spacing between rows of grid nodes.
        """
        return self._spacing[0]

    @property
    def dx(self):
        """Spacing between columns of grid nodes.
        """
        return self._spacing[1]


class RasterGrid(UniformRectilinearGrid):
    """
    Examples
    --------
    >>> from landlab.grid.structured_quad.rectilinear import RasterGrid
    >>> grid = RasterGrid((4, 5), spacing=2, origin=(-1, 1))
    >>> grid.number_of_nodes
    20

    #>>> grid.number_of_core_nodes
    6
    >>> grid.number_of_node_rows
    4
    >>> grid.number_of_node_columns
    5
    >>> grid.corner_nodes
    array([ 0,  4, 15, 19])
    >>> grid.number_of_cells
    6
    >>> grid.node_row_coord
    array([-1.,  1.,  3.,  5.])
    >>> grid.node_col_coord
    array([ 1.,  3.,  5.,  7.,  9.])
    """
    def __init__(self, shape, spacing=1., origin=0.):
        """
        Parameters
        ----------
        shape : tuple
            Shape of the grid in nodes.
        spacing : float, optional
            Spacing between rows and columns.
        origin : tuple, optional
            Coordinates of grid origin.
        """
        super(RasterGrid, self).__init__(shape, origin=origin,
                                         spacing=(spacing, ) * len(shape))



from landlab.components.nonlinear_diffusion.Perron_nl_diffuse import PerronNLDiffuse
from landlab.components.diffusion.diffusion import DiffusionComponent #...the two different diffusion formulations
from landlab import ModelParameterDictionary #handles input from the input file

from landlab import RasterModelGrid #the grid object
from landlab.plot.imshow import imshow_node_grid
import numpy as np
import pylab

#get the needed properties to build the grid:
input_file = './diffusion_test_params.txt'
#initialize an object that will supply the parameters:
inputs = ModelParameterDictionary(input_file)
nrows = inputs.read_int('nrows')
ncols = inputs.read_int('ncols')
dx = inputs.read_float('dx')
leftmost_elev = inputs.read_float('leftmost_elevation')
initial_slope = inputs.read_float('initial_slope')
uplift_rate = inputs.read_float('uplift_rate')
runtime = inputs.read_float('total_time')
dt = inputs.read_float('dt')
nt = int(runtime//dt) #"//" means "divide and truncate" ("%" means "division remainder")
uplift_per_step = uplift_rate * dt

#Instantiate the grid object
#We know which parameters are needed for input by inspecting the function where it lives, in landlab.grid.raster
#We could also look at the documentation for landlab found online (http://the-landlab.readthedocs.org)
mg = RasterModelGrid(nrows, ncols, dx)

##create the elevation field in the grid:
#create the field
mg.create_node_array_zeros('topographic_elevation')
z = mg.create_node_array_zeros() + leftmost_elev #in our case, slope is zero, so the leftmost_elev is the mean elev
#put these values plus roughness into that field
mg['node'][ 'topographic_elevation'] = z + np.random.rand(len(z))/100000.

#set up its boundary conditions (bottom, left, top, right)
#The mechanisms for this are all automated within the grid object
mg.set_fixed_value_boundaries_at_grid_edges(True, True, True, True)

# Display a message
print 'Running ...' 

#instantiate the components:
diffuse = PerronNLDiffuse(mg, input_file)
lin_diffuse = DiffusionComponent(grid=mg, input_stream=input_file)
#lin_diffuse.initialize(input_file)

#Perform the loops.
#First, we do the nonlinear diffusion:

#We're going to perform a block uplift of the interior of the grid, but leave the boundary nodes at their original elevations.
uplifted_nodes = mg.get_core_nodes() #access this function of the grid, and store the output with a local name
#(Note: Node numbering runs across from the bottom left of the grid.)

for i in xrange(nt): #nt is the number of timesteps we calculated above, i.e., loop nt times. We never actually use i within the loop, but we could do.
    #("xrange" is a clever memory-saving way of producing consecutive integers to govern a loop)
    #this colon-then-tab-in arrangement is what Python uses to delineate connected blocks of text, instead of brackets or parentheses
    #This line performs the actual functionality of the component:
    #mg = lin_diffuse.diffuse(mg, dt) #linear diffusion
    mg = diffuse.diffuse(mg, i*dt) #nonlinear diffusion
    #...swap around which line is commented out to switch between formulations of diffusion

    #now plot a N-S cross section from this stage in the run onto figure 1. The sections will all be superimposed, as show() hasn't yet been called
    pylab.figure(1)
    elev_r = mg.node_vector_to_raster(mg['node']['topographic_elevation']) #turn the 1-D array of elevation values into a spatially accurate 2-D gridded format, for plotting
    im = pylab.plot(mg.dx*np.arange(nrows), elev_r[:,int(ncols//2)]) #square brackets denote a subset of nodes to use.
    #...this kind of data extraction from a larger data structure ("slicing", or "fancy indexing") is extremely useful and powerful, and is one of the appeals of Python
    #...this is plot(x, y).
    #x is the distance north up the grid.
    #y is the elevation along all the rows, but only the 50th column (more slicing!), i.e., halfway along the grid and N-S

    print 'Completed loop ', i
 
print 'Completed the simulation. Plotting...'

#Finalize and plot:
#put a title on figure 1
pylab.figure(1)
pylab.title('N-S cross_section, nonlinear diffusion')
pylab.xlabel('Distance')
pylab.ylabel('Elevation')

#figure 2 is the map of the final elevations
pylab.figure(2)
im_nl = imshow_node_grid(mg, 'topographic_elevation')  # display a colored image

pylab.figure(3)
elev_r = mg.node_vector_to_raster(mg['node']['topographic_elevation']) #turn the 1-D array of elevation values into a spatially accurate 2-D gridded format, for plotting
im = pylab.plot(mg.dx*np.arange(nrows), elev_r[:,int(ncols//2)])

print('Done.')

#now do the linear diffusion:

##Reset the elevation field in the grid:
mg['node'][ 'topographic_elevation'] = z + np.random.rand(len(z))/100000.

# Display a message
print 'Running ...' 

##instantiate the components:
#diffuse = PerronNLDiffuse(mg, input_file)
#lin_diffuse = DiffusionComponent(grid=mg)
#lin_diffuse.initialize(input_file)

for i in xrange(nt):
    #This line performs the actual functionality of the component:
    #***NB: the nonlinear diffuser contains an "automatic" element of uplift. If you instead use the linear diffuser, you need to add the uplift manually...
    mg['node']['topographic_elevation'][uplifted_nodes] += uplift_per_step 
    mg = lin_diffuse.diffuse(mg, internal_uplift=False) #linear diffusion

    pylab.figure(4)
    elev_r = mg.node_vector_to_raster(mg['node']['topographic_elevation'])
    im = pylab.plot(mg.dx*np.arange(nrows), elev_r[:,int(ncols//2)])

    print 'Completed loop ', i

print 'Completed the simulation. Plotting...'

#Finalize and plot:
#put a title on figure 4
pylab.figure(4)
pylab.title('N-S cross_section, linear diffusion')
pylab.xlabel('Distance')
pylab.ylabel('Elevation')

#figure 5 is the map of the final elevations
pylab.figure(5)
im = imshow_node_grid(mg, 'topographic_elevation')

#superpose this final form onto figure 3:
pylab.figure(3)
elev_r = mg.node_vector_to_raster(mg['node']['topographic_elevation']) #turn the 1-D array of elevation values into a spatially accurate 2-D gridded format, for plotting
im = pylab.plot(mg.dx*np.arange(nrows), elev_r[:,int(ncols//2)])
pylab.xlabel('Distance')
pylab.ylabel('Elevation')

pylab.show() #this line displays all of the figures you've issued plot commands for, since you last called show()
